"""Repository-related errors."""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.utils.encoding import force_str
from django.utils.translation import gettext as _
from housekeeping import ClassDeprecatedMixin

from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.ssh.errors import SSHAuthenticationError

if TYPE_CHECKING:
    from reviewboard.scmtools.certs import Certificate


class SCMError(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)


class ChangeSetError(SCMError):
    pass


class InvalidChangeNumberError(ChangeSetError):
    def __init__(self):
        ChangeSetError.__init__(self, None)


class ChangeNumberInUseError(ChangeSetError):
    def __init__(self, review_request=None):
        ChangeSetError.__init__(self, None)
        self.review_request = review_request


class EmptyChangeSetError(ChangeSetError):
    def __init__(self, changenum):
        ChangeSetError.__init__(self, _('Changeset %s is empty') % changenum)


class InvalidRevisionFormatError(SCMError):
    """Indicates that a revision isn't in a recognizable format."""

    def __init__(self, path, revision, detail=None):
        """Initialize the exception.

        Args:
            path (bytes or unicode):
                The path the revision was for.

            revision (bytes or unicode):
                The revision that was invalid.

            detail (unicode, optional):
                Additional detail to display after the standard error message.
        """
        path = force_str(path)
        revision = force_str(revision)

        msg = _("The revision '%(revision)s' for '%(path)s' isn't in a valid "
                "format") % {
            'revision': revision,
            'path': path,
        }

        if detail:
            msg += ': ' + detail

        SCMError.__init__(self, msg)

        self.path = path
        self.revision = revision
        self.detail = detail


class FileNotFoundError(SCMError):
    """An error indicating a file was not found in a repository.

    Attributes:
        base_commit_id (unicode):
            The optional ID of the base commit the file and revision belonged
            to.

        context (reviewboard.scmtools.core.FileLookupContext, optional):
            Extra context used to help look up the file.

            Version Added:
                4.0.5

        detail (unicode):
            Additional details for the error message.

        path (unicode):
            The path in the repository.

        revision (reviewboard.scmtools.core.Revision or unicode):
            The revision in the repository.
    """

    def __init__(self, path, revision=None, detail=None, base_commit_id=None,
                 context=None):
        """Initialize the error.

        Args:
            path (unicode):
                The path in the repository.

            revision (reviewboard.scmtools.core.Revision or unicode, optional):
                The revision in the repository.

            detail (unicode, optional):
                Additional details for the error message.

            base_commit_id (unicode, optional):
                The optional ID of the base commit the file and revision
                belonged to.

            context (reviewboard.scmtools.core.FileLookupContext, optional):
                Extra context used to help look up the file.

                Version Added:
                    4.0.5
        """
        from reviewboard.scmtools.core import HEAD

        if isinstance(path, bytes):
            path = path.decode('utf-8', 'ignore')

        if base_commit_id is None and context is not None:
            base_commit_id = context.base_commit_id

        if revision is None or revision == HEAD and base_commit_id is None:
            msg = (_("The file '%s' could not be found in the repository")
                   % path)
        elif base_commit_id is not None and base_commit_id != revision:
            msg = _('The file "%(path)s" (revision %(revision)s, commit '
                    '%(base_commit_id)s) could not be found in the '
                    'repository') % {
                'path': path,
                'revision': revision,
                'base_commit_id': base_commit_id,
            }
        else:
            msg = _('The file "%(path)s" (revision %(revision)s) could not be '
                    'found in the repository') % {
                'path': path,
                'revision': revision,
            }

        if detail:
            msg += ': ' + detail

        Exception.__init__(self, msg)

        self.revision = revision
        self.base_commit_id = base_commit_id
        self.context = context
        self.path = path
        self.detail = detail


class RepositoryNotFoundError(SCMError):
    """An error indicating that a given path is not a valid repository.

    Version Changed:
        4.0.11:
        Added :py:attr:`form_field_id` and an equivalent argument to the
        constructor.

    Attributes:
        form_field_id (unicode):
            The ID of the form field that this error corresponds to.

            This may be ``None`` if this error isn't about a specific field.

            Version Added:
                4.0.11
    """

    def __init__(self, msg=None, form_field_id=None):
        """Initialize the error.

        Version Changed:
            4.0.11:
            This now takes optional ``msg`` and ``form_field_id`` parameters.

        Args:
            msg (unicode, optional):
                The optional custom message to display.

            form_field_id (unicode, optional):
                The optional ID of the form field that this error corresponds
                to.
        """
        super(RepositoryNotFoundError, self).__init__(
            msg or
            _('A repository was not found. Please check the configuration '
              'to make sure the details are correct and that it allows '
              'access using any credentials you may have provided.'))

        self.form_field_id = form_field_id


class AuthenticationError(SSHAuthenticationError, SCMError):
    """An error representing a failed authentication for a repository.

    This takes a list of authentication types that are allowed. These
    are dependent on the backend, but are loosely based on SSH authentication
    mechanisms. Primarily, we respond to "password" and "publickey".

    This may also take the user's SSH key that was tried, if any.
    """
    pass


class UnverifiedCertificateError(
    ClassDeprecatedMixin,
    SCMError,
    warning_cls=RemovedInReviewBoard90Warning,
    init_deprecation_msg=(
        'UnverifiedCertificateError is deprecated in favor of '
        'reviewboard.certs.errors.CertificateVerificationError, and '
        'will be removed in Review Board 9.'
    )
):
    """An error representing an unverified SSL certificate.

    Deprecated:
        6.0:
        This is deprecated in favor of
        :py:class:`reviewboard.certs.errors.CertificateVerificationError`,
        and will be removed in Review Board 9.
    """

    ######################
    # Instance variables #
    ######################

    #: The certificate generating this error.
    #:
    #: Type:
    #:     reviewboard.scmtools.certs.Certificate
    certificate: Certificate

    def __init__(
        self,
        certificate: Certificate,
    ) -> None:
        """Initialize the error message.

        Args:
            certificate (reviewboard.scmtools.certs.Certificate):
                The certificate this error pertains to.
        """
        info = []

        if certificate.hostname:
            info.append(_('hostname "%s"') % certificate.hostname)

        if certificate.fingerprint:
            info.append(_('fingerprint "%s"') % certificate.fingerprint)

        if certificate and certificate.fingerprint:
            msg = _(
                'The SSL certificate for this repository (%s) was not '
                'verified and might not be safe. This certificate needs to '
                'be verified before the repository can be accessed.'
            ) % (', '.join(info))
        else:
            msg = _(
                'The SSL certificate for this repository was not verified '
                'and might not be safe. This certificate needs to be '
                'verified before the repository can be accessed.'
            )

        super().__init__(msg)
        self.certificate = certificate



"""Unit tests for reviewboard.reviews.ui.image.ImageReviewUI."""

from djblets.util.templatetags.djblets_images import crop_image

from reviewboard.admin.server import build_server_url
from reviewboard.reviews.ui.image import ImageReviewUI
from reviewboard.testing import TestCase


class ImageReviewUITests(TestCase):
    """Unit tests for reviewboard.reviews.ui.image.ImageReviewUI."""

    fixtures = ['test_users']

    def setUp(self):
        super(ImageReviewUITests, self).setUp()

        self.review_request = self.create_review_request()
        self.attachment = self.create_file_attachment(
            self.review_request)
        self.review = self.create_review(self.review_request)

    def test_get_comment_thumbnail(self):
        """Testing ImageReviewUI.get_comment_thumbnail for an image comment"""
        ui = ImageReviewUI(self.review_request, self.attachment)
        comment = self.create_file_attachment_comment(
            self.review,
            self.attachment,
            extra_fields={
                'x': 0,
                'y': 0,
                'width': 1,
                'height': 1,
            })
        thumbnail = ui.get_comment_thumbnail(comment)

        self.assertHTMLEqual(
            thumbnail,
            '<img class="modified-image" src="%s" width="1" height="1"'
            ' alt="%s" />'
            % (build_server_url(crop_image(self.attachment.file, 0, 0, 1, 1)),
               comment.text)
        )

    def test_get_comment_thumbnail_diff(self):
        """Testing ImageReviewUI.get_comment_thumbnail for an image diff
        comment
        """
        diff_attachment = self.create_file_attachment(self.review_request)

        ui = ImageReviewUI(self.review_request, self.attachment)
        ui.set_diff_against(diff_attachment)

        comment = self.create_file_attachment_comment(
            self.review,
            self.attachment,
            diff_attachment,
            extra_fields={
                'x': 0,
                'y': 0,
                'width': 1,
                'height': 1,
            })
        thumbnail = ui.get_comment_thumbnail(comment)

        self.assertHTMLEqual(
            thumbnail,
            '<div class="image-review-ui-diff-thumbnail">'
            '<img class="orig-image" src="%s" width="1" height="1" alt="%s" />'
            '<img class="modified-image" src="%s" width="1" height="1"'
            ' alt="%s" />'
            '</div>'
            % (build_server_url(crop_image(diff_attachment.file, 0, 0, 1, 1)),
               comment.text,
               build_server_url(crop_image(self.attachment.file, 0, 0, 1, 1)),
               comment.text)
        )



"""NIS authentication backend."""

from __future__ import annotations

import logging
from importlib import import_module
from types import ModuleType
from typing import Optional, TYPE_CHECKING, Tuple

from django.conf import settings
from django.contrib.auth.models import User
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from typing_extensions import TypeAlias

from reviewboard.accounts.backends.base import BaseAuthBackend
from reviewboard.accounts.forms.auth import NISSettingsForm

if TYPE_CHECKING:
    from django.http import HttpRequest


logger = logging.getLogger(__name__)


#: A type alias for a NIS password entry.
#:
#: Version Added:
#:     6.0
_NISPasswdEntry: TypeAlias = Tuple[str, str, int, int, str, str, str]


class NISBackend(BaseAuthBackend):
    """Authenticate against a user on an NIS server."""

    backend_id = 'nis'
    name = _('NIS')
    settings_form = NISSettingsForm
    login_instructions = \
        _('Use your standard NIS username and password.')

    @cached_property
    def nis(self) -> Optional[ModuleType]:
        """The nis module, used for interacting with NIS.

        On first access, this will check if NIS is available, logging an
        error if missing.

        This safeguards against Python environments without NIS, and against
        versions of Python >= 3.13.

        Type:
            module
        """
        try:
            return import_module('nis')
        except ImportError:
            logger.error('The nis module is not available on your version '
                         'of Python.')
            return None

    def authenticate(
        self,
        request: Optional[HttpRequest] = None,
        *,
        username: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs,
    ) -> Optional[User]:
        """Authenticate the user.

        This will attempt to authenticate the user against NIS. If the
        username and password are valid, a user will be returned, and added
        to the database if it doesn't already exist.

        Version Changed:
            6.0:
            * ``request`` is now optional.
            * ``username`` and ``password`` are technically optional, to
              aid in consistency for type hints, but will result in a ``None``
              result.

        Version Changed:
            4.0:
            The ``request`` argument is now mandatory as the first positional
            argument, as per requirements in Django.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the caller. This may be ``None``.

            username (str):
                The username to authenticate.

            password (str):
                The password to authenticate.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            django.contrib.auth.models.User:
            The authenticated user, or ``None`` if the user could not be
            authenticated.
        """
        if not username or not password:
            # This may be an authentication request for a backend expecting
            # different arguments.
            return None

        nis = self.nis

        if nis is None:
            return None

        username = username.strip()

        import crypt

        try:
            passwd = self._nis_get_passwd(username)
        except nis.error:
            # The user does not exist, or there was an error communicating
            # with NIS.
            return None

        original_crypted = passwd[1]
        new_crypted = crypt.crypt(password, original_crypted)

        if original_crypted == new_crypted:
            return self.get_or_create_user(username=username,
                                           request=request,
                                           passwd=passwd)

        return None

    def get_or_create_user(
        self,
        username: str,
        request: Optional[HttpRequest] = None,
        passwd: Optional[_NISPasswdEntry] = None,
    ) -> Optional[User]:
        """Return an existing user, or create one if it does not exist.

        If the user does not exist in the database, but does in the backend,
        its information will be stored in the database for later lookup.

        Args:
            username (str):
                The username of the user.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

            passwd (tuple, optional):
                The parsed NIS passwd entry for the user.

        Returns:
            django.contrib.auth.models.User:
            The existing or newly-created user, or ``None`` if an error was
            encountered.
        """
        nis = self.nis

        if nis is None:
            return None

        username = username.strip()

        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            if not passwd:
                try:
                    passwd = self._nis_get_passwd(username)
                except nis.error:
                    # The user does not exist, or there was an error
                    # communicating with NIS.
                    return None

            names = passwd[4].split(',')[0].split(' ', 1)
            first_name = names[0]
            last_name = None

            if len(names) > 1:
                last_name = names[1]

            email = '%s@%s' % (username, settings.NIS_EMAIL_DOMAIN)

            user = User(username=username,
                        password='',
                        first_name=first_name,
                        last_name=last_name or '',
                        email=email)
            user.is_staff = False
            user.is_superuser = False
            user.set_unusable_password()
            user.save()

        return user

    def _nis_get_passwd(
        self,
        username: str,
    ) -> _NISPasswdEntry:
        """Return a passwd entry for a user.

        Args:
            username (str):
                The username to fetch.

        Returns:
            tuple:
            The parsed passwd entry.

        Raises:
            nis.error:
                The user could not be found, or there was an error performing
                the lookup.
        """
        nis = self.nis
        assert nis

        return nis.match(username, 'passwd').split(':')



"""Functions for retrieving server information."""

from __future__ import annotations

import os
import socket
from typing import Optional
from urllib.parse import urljoin

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext as _
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.site.urlresolvers import local_site_reverse


#: A cached path containing the site's data directory.
_data_dir: Optional[str] = None


def get_server_url(local_site_name=None, local_site=None, request=None):
    """Return the URL for the root of the server.

    This will construct a URL that points to the root of the server, factoring
    in whether to use HTTP or HTTPS.

    If ``local_site_name`` or ``local_site`` is provided, then the URL will be
    the root to the LocalSite's root, rather than the server's root.

    If ``request`` is provided, then the Local Site, if any, will be
    inferred from the request.
    """
    site = Site.objects.get_current()
    siteconfig = SiteConfiguration.objects.get_current()
    root = local_site_reverse('root', local_site_name=local_site_name,
                              local_site=local_site, request=request)

    return '%s://%s%s' % (siteconfig.get('site_domain_method'),
                          site.domain, root)


def build_server_url(path=None, **kwargs):
    """Build an absolute URL containing the full URL to the server.

    A path can be supplied that will be joined to the server URL.

    Args:
        path (unicode):
            The path to append to the server URL.

        **kwargs (dict):
            Additional arguments to pass to :py:func:`get_server_url`.

    Returns:
        unicode:
        The resulting URL.
    """
    return urljoin(get_server_url(**kwargs), path)


def get_hostname():
    """Return the hostname for this Review Board server.

    Returns:
        unicode:
        The hostname for the server.
    """
    return str(socket.gethostname())


def get_data_dir() -> str:
    """Return the path to the site's data directory.

    This is always based on :envvar:`$HOME`. If this variable is not set,
    or the path does not exist, then an exception will be raised.

    Version Added:
        6.0

    Returns:
        str:
        The path to the data directory.

    Raises:
        django.core.exceptions.ImproperlyConfigured:
            The data directory path could not be found or does not exist.

            Details are in the error message.
    """
    global _data_dir

    if _data_dir != settings.SITE_DATA_DIR:
        _data_dir = settings.SITE_DATA_DIR

        if not _data_dir:
            raise ImproperlyConfigured(
                _('The site data directory could not be determined. '
                  'Please make sure your web server is using our '
                  'provided reviewboard.wsgi module for WSGI.'))

        if not os.path.exists(_data_dir):
            raise ImproperlyConfigured(
                _('The site data directory (%s) does not exist. Please '
                  'make sure you are running in the right environment with a '
                  'working site directory.') % _data_dir)

    return _data_dir



"""Evolution to move away from MultiSelectField.

Version Added:
    7.0
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('WebHookTarget', 'events', initial=None, max_length=512),
]



"""Client for Forgejo.

Version Added:
    8.0
"""

from __future__ import annotations

import json
import logging
import re
from base64 import b64decode
from typing import List, TypeVar, TYPE_CHECKING
from urllib.error import HTTPError
from urllib.parse import quote as urlquote, urlencode
from uuid import uuid4

from django.utils.translation import gettext as _
from pydantic import BaseModel, TypeAdapter, ValidationError

from reviewboard.hostingsvcs.base.client import HostingServiceClient
from reviewboard.hostingsvcs.errors import (
    AuthorizationError,
    HostingServiceError,
    HostingServiceAPIError,
    TwoFactorAuthCodeRequiredError,
)
from reviewboard.hostingsvcs.forgejo import api
from reviewboard.hostingsvcs.forgejo.errors import APITokenNameInUseError
from reviewboard.scmtools.core import Branch, Commit
from reviewboard.scmtools.crypto_utils import decrypt_password
from reviewboard.scmtools.errors import (
    FileNotFoundError,
    RepositoryNotFoundError,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from reviewboard.hostingsvcs.base.bug_tracker import BugInfo
    from reviewboard.hostingsvcs.base.hosting_service import \
        HostingServiceCredentials
    from reviewboard.hostingsvcs.forgejo.service import Forgejo
    from reviewboard.hostingsvcs.models import HostingServiceAccount
    from reviewboard.scmtools.models import Repository

    _T = TypeVar('_T')
    _TModel = TypeVar('_TModel', bound=BaseModel)


logger = logging.getLogger(__name__)


class ForgejoClient(HostingServiceClient):
    """Hosting service client for Forgejo.

    Version Added:
        8.0
    """

    #: Regex for index lines in the diff.
    _index_line_re: re.Pattern[bytes]

    #: Regex for SHAs for deleted/added files.
    _missing_file_sha_re: re.Pattern[str]

    def __init__(
        self,
        hosting_service: Forgejo,
    ) -> None:
        """Initialize the client.

        Args:
            hosting_service (Forgejo):
                The hosting service instance.
        """
        super().__init__(hosting_service)

        self.account = hosting_service.account
        self._index_line_re = re.compile(
            br'^index (?P<a>[0-9a-f]+)..(?P<b>[0-9a-f]+)((?P<rest>\s.*)?)$')
        self._missing_file_sha_re = re.compile(r'^0+$')

    def get_http_credentials(
        self,
        account: HostingServiceAccount,
        username: (str | None) = None,
        password: (str | None) = None,
        **kwargs,
    ) -> HostingServiceCredentials:
        """Return credentials used to authenticate with Forgejo.

        Args:
            account (reviewboard.hostingsvcs.models.HostingServiceAccount):
                The stored authentication data for the service.

            username (str, optional):
                An explicit username passed by the caller. This is not used.

            password (str, optional):
                An explicit password passed by the caller. This is not used.

            **kwargs (dict, unused):
                Additional keyword arguments passed in when making the HTTP
                request.

        Returns:
            dict:
            A dictionary of credentials for the request.
        """
        if encrypted_token := account.data.get('api_token'):
            token = decrypt_password(encrypted_token)

            return {
                'headers': {
                    'Authorization': f'token {token}',
                },
            }

        return super().get_http_credentials(
            account, username, password, **kwargs)

    def check_repository(
        self,
        *,
        hosting_url: str,
        repository_owner: str,
        repository_name: str,
    ) -> None:
        """Check that a repository is valid and accessible.

        Args:
            hosting_url (str):
                The URL of the Forgejo server.

            repository_owner (str):
                The repository owner.

            repository_name (str):
                The repository name.

        Raises:
            reviewboard.scmtools.errors.RepositoryNotFoundError:
                A repository was not found with the given parameters.
        """
        url = self._get_api_repo_root(
            hosting_url=hosting_url,
            repository_owner=repository_owner,
            repository_name=repository_name)

        try:
            # We don't actually use any fields from this (right now), but we
            # want to attempt to deserialize just to verify that we have a
            # valid repository response.
            self._api_get(
                url=url,
                result_type=api.Repository)
        except HostingServiceAPIError as e:
            error = e.rsp
            assert isinstance(error, api.APIError)

            if error.message == 'GetUserByName':
                raise RepositoryNotFoundError(
                    (
                        _('A user with the name {repository_owner} was not '
                          'found or is not accessible.')
                        .format(repository_owner=repository_owner)
                    ),
                    form_field_id='repository_owner')
            elif e.http_code == 404:
                raise RepositoryNotFoundError(
                    (
                        _('A repository with the name {repository_name} was '
                          'not found for the owner {repository_owner}')
                        .format(repository_owner=repository_owner,
                                repository_name=repository_name)
                    ),
                    form_field_id='repository_name')
            else:
                logger.warning(
                    'Unexpected error when checking repository with API URL '
                    '%s: %s',
                    url, str(e))

                raise RepositoryNotFoundError()
        except HostingServiceError as e:
            logger.warning(
                'Unexpected error when checking repository with API URL '
                '%s: %s',
                url, str(e))

            raise RepositoryNotFoundError()

    def create_api_token(
        self,
        *,
        token_name: str,
        hosting_url: str,
        username: str,
        password: str,
        two_factor_auth_code: (str | None) = None,
    ) -> str:
        """Create an API token for the given user.

        Args:
            token_name (str):
                The name to use for the token.

            hosting_url (str):
                The URL of the Forgejo server.

            username (str):
                The username for the account.

            password (str):
                The password for the account.

            two_factor_auth_code (str, optional):
                The two-factor authentication code provided by the user.

        Returns:
            str:

        Raises:
            reviewboard.certs.errors.CertificateVerificationError:
                The certificate for the host is invalid or untrusted.

            reviewboard.hostingsvcs.errors.AuthorizationError:
                The credentials provided were not valid.

            reviewboard.hostingsvcs.errors.TwoFactorAuthCodeRequiredError:
                A two-factor authentication code is required to authorize
                this account. The request must be retried with the same
                credentials and with the ``two_factor_auth_code`` parameter
                provided.

            reviewboard.hostingsvcs.errors.HostingServiceError:
                A non-authentication error occurred.

            reviewboard.hostingsvcs.forgejo.errors.APITokenNameInUseError:
                The API token name was already used.
        """
        api_root = self._get_api_root(hosting_url)
        url_username = urlquote(username)
        url = f'{api_root}/users/{url_username}/tokens'
        trace_id = str(uuid4())

        # Temporarily null-out any existing tokens, so we don't try to use
        # it if we're reauthenticating ("Edit Credentials")
        saved_token = self.account.data.get('api_token')

        try:
            del self.account.data['api_token']
        except Exception:
            pass

        try:
            body = json.dumps({
                'name': token_name,
                'scopes': [
                    'read:issue',
                    'read:organization',
                    'read:repository',
                    'read:user',
                ],
            }).encode()

            headers = {
                'Accept': 'application/json',
            }

            if two_factor_auth_code:
                headers['X-Forgejo-OTP'] = two_factor_auth_code

            rsp = self.http_post(
                url,
                body=body,
                content_type='application/json',
                headers=headers,
                username=username,
                password=password)

            logger.info('Created API token for user %s on server %s',
                        username, hosting_url)
            access_token = api.AccessToken.model_validate_json(rsp.data)

            return access_token.sha1
        except ValidationError as e:
            logger.error('[%s] Data validation failed for API token creation: '
                         '%s',
                         trace_id, e)

            if saved_token:
                self.account.data['api_token'] = saved_token

            raise HostingServiceError(
                _(
                    'Unexpected response from Forgejo server. Check the '
                    'Review Board server logs for more information (error ID '
                    '{trace_id}'
                ).format(trace_id=trace_id))
        except HTTPError as e:
            data = e.read()

            logger.error('[%s] HTTP error (code=%s) creating API token with '
                         'POST %s: %s ',
                         trace_id, e.code, url, data.decode())

            if saved_token:
                self.account.data['api_token'] = saved_token

            try:
                error = api.APIError.model_validate_json(data)
            except ValidationError:
                error = data.decode()
                error_code = e.code

                logger.error(
                    '[%s] Unexpected error (HTTP %d) when creating API token '
                    'for user %s on server %s: %s',
                    trace_id, error_code, username, hosting_url, error)

                raise HostingServiceError(
                    _(
                        'Unknown error while creating Forgejo API token. '
                        'See the Review Board server logs for more '
                        'information (error ID {trace_id}).'
                    )
                    .format(trace_id=trace_id))

            if error.message == 'access token name has been used already':
                logger.warning(
                    'Creating API token (name="%s") for user %s on server %s '
                    'failed due to duplicate name.',
                    token_name, username, hosting_url)

                raise APITokenNameInUseError()
            elif error.message == 'invalid provided OTP':
                logger.info('Invalid OTP when creating API token for user %s '
                            'on server %s',
                            username, hosting_url)

                raise TwoFactorAuthCodeRequiredError(
                    _('Enter your two-factor authentication code'))
            elif e.code == 401:
                logger.warning('Authentication failure when creating API '
                               'token for user %s on server %s',
                               username, hosting_url)

                raise AuthorizationError(
                    _('The username or password is incorrect.'))
            else:
                logger.error('Unexpected error when creating API token for '
                             'user %s on server %s: %s',
                             username, hosting_url, error.message)

                raise HostingServiceError(error.message)

    def get_blob(
        self,
        *,
        hosting_url: str,
        repository: Repository,
        path: str,
        sha: str,
    ) -> bytes:
        """Retrieve a git blob.

        Args:
            hosting_url (str):
                The URL of the Forgejo server.

            repository (reviewboard.scmtools.models.Repsitory):
                The repository.

            path (str):
                The path of the file.

            sha (str):
                The hash of the blob to retrieve.

        Returns:
            bytes:
            The contents of the file.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                An error occurred when fetching the file.

            reviewboard.scmtools.errors.FileNotFoundError:
                The blob with the given SHA was not found.
        """
        repo_root = self._get_api_repo_root(hosting_url, repository)

        try:
            url_sha = urlquote(sha)
            blob = self._api_get(
                url=f'{repo_root}/git/blobs/{url_sha}',
                result_type=api.GitBlobResponse,
                repository=repository)

            if blob.encoding != 'base64':
                logger.error('Failed to get blob "%s" for repository pk=%d: '
                             'encoding was "%s" but only base64 is '
                             'supported.',
                             sha, repository.pk, blob.encoding)

                raise HostingServiceError(
                    _('Forgejo returned data with an unknown encoding.'))

            return b64decode(blob.content)
        except HostingServiceAPIError as e:
            error = e.rsp
            assert isinstance(error, api.APIError)

            # If a blob with the given SHA cannot be found, this endpoint
            # returns HTTP 400 with this message. We only get HTTP 404 if
            # the owner/repository names in the URL are incorrect.
            if error.message.startswith('object does not exist'):
                raise FileNotFoundError(path, revision=sha)
            else:
                raise e
        except Exception as e:
            raise HostingServiceError(str(e))

    def get_branches(
        self,
        *,
        hosting_url: str,
        repository: Repository,
    ) -> Sequence[Branch]:
        """Return a list of all branches in the repositories.

        Args:
            hosting_url (str):
                The URL of the Forgejo server.

            repository (reviewboard.scmtools.models.Repository):
                The repository for which branches should be returned.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            The branches.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error fetching branches.
        """
        repo_root = self._get_api_repo_root(hosting_url, repository)

        repo_info = self._api_get(url=repo_root, result_type=api.Repository)

        refs = self._api_get(
            url=f'{repo_root}/git/refs',
            result_type=TypeAdapter(List[api.GitReference]),
            repository=repository)

        results: list[Branch] = []

        default_ref: (int | None) = None
        main_ref: (int | None) = None
        master_ref: (int | None) = None

        for i, ref in enumerate(refs):
            if ref.ref.startswith('refs/heads/'):
                name = ref.ref[len('refs/heads/'):]
            else:
                name = ref.ref

            results.append(Branch(
                id=name,
                commit=ref.object.sha,
            ))

            if name == repo_info.default_branch:
                default_ref = i
            if name == 'main':
                main_ref = i
            elif name == 'master':
                master_ref = i

        # Prefer the configured default, then "main", then "master".
        if default_ref is not None:
            results[default_ref].default = True
        elif main_ref is not None:
            results[main_ref].default = True
        elif master_ref is not None:
            results[master_ref].default = True
        elif len(results) > 0:
            results[0].default = True

        return results

    def get_commits(
        self,
        hosting_url: str,
        repository: Repository,
        branch: (str | None) = None,
        start: (str | None) = None,
    ) -> Sequence[Commit]:
        """Return a list of commits backward in history from a given point.

        Args:
            hosting_url (str):
                The URL of the Forgejo server.

            repository (reviewboard.scmtools.models.Repository):
                The repository to retrieve commits from.

            branch (str, optional):
                The branch to retrieve from.

                If this is not provided, the default branch will be used.

            start (str, optional):
                An optional starting revision.

                If this is not provided, the most recent commits will be
                returned.

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The retrieved commits.

        Raises:
            ValueError:
                Neither a ``branch`` or ``start`` parameter was provided.

            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error fetching commits.
        """
        repo_root = self._get_api_repo_root(hosting_url, repository)

        sha = start or branch

        if not sha:
            raise ValueError('Either "branch" or "start" must be provided.')

        assert isinstance(sha, str)

        commits = self._api_get(
            url=f'{repo_root}/commits',
            result_type=TypeAdapter(List[api.Commit]),
            params={
                'sha': sha,
                'stat': '0',
                'verification': '0',
                'files': '0',
                'limit': '30',
            },
            repository=repository)

        results: list[Commit] = []

        for commit in commits:
            if commit.parents:
                parent = commit.parents[0].sha
            else:
                parent = ''

            results.append(Commit(
                author_name=commit.commit.author.name,
                id=commit.sha,
                date=commit.created,
                message=commit.commit.message,
                parent=parent,
            ))

        return results

    def get_change(
        self,
        hosting_url: str,
        repository: Repository,
        revision: str,
    ) -> Commit:
        """Return an individual change.

        Args:
            hosting_url (str):
                The URL of the Forgejo server.

            repository (reviewboard.scmtools.models.Repository):
                The repository to get the change from.

            revision (str):
                The revision to retrieve.

        Returns:
            reviewboard.scmtools.core.Commit:
            The change.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error fetching the commit.
        """
        repo_root = self._get_api_repo_root(hosting_url, repository)
        url_revision = urlquote(revision)

        commit = self._api_get(
            url=f'{repo_root}/git/commits/{url_revision}',
            result_type=api.Commit,
            repository=repository)

        if commit.parents:
            parent = commit.parents[0].sha
        else:
            parent = ''

        diff_url = f'{repo_root}/git/commits/{url_revision}.diff'

        try:
            diff = self.http_get(diff_url).data.splitlines()
        except HTTPError as e:
            logger.error('HTTP error fetching diff at %s (repository %s): %s',
                         diff_url, repository, e)

            raise HostingServiceError(
                'Failed to fetch diff data from Forgejo.')

        all_shas = (
            self._get_blob_shas_in_tree(repo_root, commit.sha, repository) |
            self._get_blob_shas_in_tree(repo_root, parent, repository))

        diff_lines: list[bytes] = []
        index_line_re = self._index_line_re
        missing_file_sha_re = self._missing_file_sha_re

        def expand_sha(
            partial_sha: str,
        ) -> str:
            if missing_file_sha_re.match(partial_sha):
                return '0' * 40

            for full_sha in all_shas:
                if full_sha.startswith(partial_sha):
                    return full_sha

            logger.warning('Unable to find full SHA for blob %s (repository '
                           '%s)',
                           partial_sha, repository)
            return partial_sha

        for line in diff:
            m = index_line_re.match(line)

            if m:
                sha1 = expand_sha(m.group('a').decode())
                sha2 = expand_sha(m.group('b').decode())
                rest = m.group('rest') or b''

                new_line = (b'index %s..%s%s'
                            % (sha1.encode(), sha2.encode(), rest))

                diff_lines.append(new_line)
            else:
                new_line = line
                diff_lines.append(line)

        return Commit(
            author_name=commit.commit.author.name,
            id=commit.sha,
            date=commit.created,
            message=commit.commit.message,
            parent=parent,
            diff=b'\n'.join(diff_lines),
        )

    def get_bug_info(
        self,
        hosting_url: str,
        repository: Repository,
        bug_id: str,
    ) -> BugInfo:
        """Return the information for the specified bug.

        Args:
            hosting_url (str):
                The URL of the Forgejo server.

            repository (reviewboard.scmtools.models.Repository):
                The repository object.

            bug_id (str):
                The ID of the bug to fetch.

        Returns:
            reviewboard.hostingsvcs.bugtracker.BugInfo:
            Information about the bug.
        """
        repo_root = self._get_api_repo_root(hosting_url, repository)

        result: BugInfo = {
            'summary': '',
            'description': '',
            'status': '',
        }

        try:
            url_bug_id = urlquote(bug_id)
            issue = self._api_get(
                url=f'{repo_root}/issues/{url_bug_id}',
                result_type=api.Issue,
                repository=repository)

            result['summary'] = issue.title
            result['description'] = issue.body
            result['description_text_format'] = 'markdown'
            result['status'] = issue.state
        except HostingServiceError:
            # Errors will already be logged, and the failure mode for this is
            # to return the result with empty strings.
            pass

        return result

    def _api_get(
        self,
        *,
        url: str,
        result_type: type[_TModel] | TypeAdapter[_T],
        params: (dict[str, str] | None) = None,
        repository: (Repository | None) = None,
    ) -> _TModel | _T:
        """Perform a GET request to the API.

        Args:
            url (str):
                The URL of the API endpoint.

            result_type (pydantic.BaseModel or pydantic.TypeAdapter):
                The pydantic model or adapter to use for deserialization.

            params (dict):
                Parameters to include in the URL.

            repository (reviewboard.scmtools.models.Repository, optional):
                The repository object, if available.

        Returns:
            object:
            The deserialized data.

        Raises:
            reviewboard.hostingsvcs.base.errors.HostingServiceError:
                An error occurred while making the request.

            reviewboard.hostingsvcs.base.errors.HostingServiceAPIError:
                An error occurred while making the request, with a parsed error
                structure.
        """
        if params:
            query = urlencode(params)
            url = f'{url}?{query}'

        logger.debug('Making GET request to %s', url)

        try:
            rsp = self.http_get(url)

            if isinstance(result_type, TypeAdapter):
                return result_type.validate_json(rsp.data)
            else:
                return result_type.model_validate_json(rsp.data)
        except ValidationError as e:
            logger.error('Data validation failed for API GET %s '
                         '(repository=%s): %s',
                         url, repository, e)

            raise HostingServiceError(
                _('Unexpected response from Forgejo server.'))
        except HTTPError as e:
            logger.error('HTTP error in API GET %s for repository %s: %s',
                         url, repository, e)

            data = e.read()

            try:
                error = api.APIError.model_validate_json(data)

                raise HostingServiceAPIError(
                    _('API error from Forgejo: {e}').format(e=error.message),
                    http_code=e.code,
                    rsp=error,
                )
            except ValidationError as pydantic_err:
                logger.error('Unable to parse error response for API GET %s '
                             '(repository=%s): %s',
                             url, repository, pydantic_err)

                raise HostingServiceError(
                    _('Unknown response from Forgejo: {rsp}').format(
                        rsp=data.decode()),
                    http_code=e.code)

    def _get_api_root(
        self,
        hosting_url: str,
    ) -> str:
        """Return the root URL for the API.

        Args:
            hosting_url (str):
                The URL of the Forgejo server.

        Returns:
            str:
            The base URL for the API.
        """
        if not hosting_url.endswith('/'):
            hosting_url += '/'

        return f'{hosting_url}api/v1'

    def _get_api_repo_root(
        self,
        hosting_url: str,
        repository: (Repository | None) = None,
        repository_owner: (str | None) = None,
        repository_name: (str | None) = None,
    ) -> str:
        """Get the API root for a repository in the API.

        Args:
            hosting_url (str):
                The URL of the Forgejo server.

            repository (reviewboard.scmtools.models.Repository, optional):
                The repository object.

            repository_owner (str, optional):
                The repository owner.

            repository_name (str, optional):
                The repository name.

        Returns:
            str:
            The base URL for API resources for the repository.
        """
        if repository:
            repository_owner = urlquote(
                repository.extra_data['repository_owner'])
            repository_name = urlquote(
                repository.extra_data['repository_name'])
        elif not repository_owner or not repository_name:
            raise ValueError('Either a repository or an owner/name must be '
                             'provided')

        api_root = self._get_api_root(hosting_url)

        return f'{api_root}/repos/{repository_owner}/{repository_name}'

    def _get_blob_shas_in_tree(
        self,
        repo_root: str,
        commit: str,
        repository: Repository,
    ) -> set[str]:
        """Return all the SHAs in the tree for a given commit.

        Args:
            repo_root (str):
                The URL root for the repository.

            commit (str):
                The SHA of the commit to fetch the tree for.

            repository (reviewboard.scmtools.models.Repository):
                The repository.

        Returns:
            set:
            A set of all the blob SHAs in the tree.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                There was an error fetching the tree.
        """
        url_commit = urlquote(commit)
        url = f'{repo_root}/git/trees/{url_commit}'
        result: set[str] = set()

        page = 1

        while 1:
            tree_page = self._api_get(
                url=url,
                result_type=api.GitTreeResponse,
                params={
                    'page': str(page),
                    'recursive': 'true',
                },
                repository=repository)

            if tree_page.tree is None:
                break

            for entry in tree_page.tree:
                result.add(entry.sha)

            if tree_page.truncated:
                page += 1
            else:
                break

        return result



"""Unit tests for mercurial."""

from __future__ import annotations

import json
import os
import unittest
from typing import Optional, TYPE_CHECKING

import kgb
from djblets.testing.decorators import add_fixtures

from reviewboard.diffviewer.testing.mixins import DiffParserTestingMixin
from reviewboard.scmtools.core import (
    FileLookupContext,
    HEAD,
    PRE_CREATION,
    Revision,
)
from reviewboard.scmtools.errors import SCMError, FileNotFoundError
from reviewboard.scmtools.hg import (HgDiffParser,
                                     HgGitDiffParser,
                                     HgTool,
                                     HgWebClient)
from reviewboard.scmtools.tests.testcases import SCMTestCase
from reviewboard.testing import online_only
from reviewboard.testing.testcase import TestCase

if TYPE_CHECKING:
    from typelets.django.json import SerializableDjangoJSONDict

    from reviewboard.scmtools.core import RevisionID, SCMClient


class MercurialTests(DiffParserTestingMixin, SCMTestCase):
    """Unit tests for mercurial."""

    fixtures = ['test_scmtools']

    def setUp(self) -> None:
        """Set up the test."""
        super().setUp()

        hg_repo_path = os.path.join(os.path.dirname(__file__),
                                    '..', 'testdata', 'hg_repo')
        self.repository = self.create_repository(
            name='Test HG',
            path=hg_repo_path,
            tool_name='Mercurial')

        try:
            self.tool = self.repository.get_scmtool()
        except ImportError:
            raise unittest.SkipTest('Hg is not installed')

    def test_ssh_disallowed(self) -> None:
        """Testing HgTool does not allow SSH URLs"""
        with self.assertRaises(SCMError):
            self.tool.check_repository('ssh://foo')

    def test_git_parser_selection_with_header(self) -> None:
        """Testing HgTool returns the git parser when a header is present"""
        diffContents = (b'# HG changeset patch\n'
                        b'# Node ID 6187592a72d7\n'
                        b'# Parent  9d3f4147f294\n'
                        b'diff --git a/emptyfile b/emptyfile\n'
                        b'new file mode 100644\n')

        parser = self.tool.get_parser(diffContents)
        self.assertEqual(type(parser), HgGitDiffParser)

    def test_hg_parser_selection_with_header(self) -> None:
        """Testing HgTool returns the hg parser when a header is present"""
        diffContents = (b'# HG changeset patch'
                        b'# Node ID 6187592a72d7\n'
                        b'# Parent  9d3f4147f294\n'
                        b'diff -r 9d3f4147f294 -r 6187592a72d7 new.py\n'
                        b'--- /dev/null   Thu Jan 01 00:00:00 1970 +0000\n'
                        b'+++ b/new.py  Tue Apr 21 12:20:05 2015 -0400\n')

        parser = self.tool.get_parser(diffContents)
        self.assertEqual(type(parser), HgDiffParser)

    def test_git_parser_sets_commit_ids(self) -> None:
        """Testing HgGitDiffParser sets the parser commit ids"""
        diffContents = (b'# HG changeset patch\n'
                        b'# Node ID 6187592a72d7\n'
                        b'# Parent  9d3f4147f294\n'
                        b'diff --git a/emptyfile b/emptyfile\n'
                        b'new file mode 100644\n')

        parser = self.tool.get_parser(diffContents)
        parser.parse()
        self.assertEqual(parser.new_commit_id, b'6187592a72d7')
        self.assertEqual(parser.base_commit_id, b'9d3f4147f294')

    def test_patch_creates_new_file(self) -> None:
        """Testing HgTool with a patch that creates a new file"""
        self.assertEqual(
            self.tool.parse_diff_revision(filename=b'/dev/null',
                                          revision=b'bf544ea505f8')[1],
            PRE_CREATION)

    def test_diff_parser_new_file(self) -> None:
        """Testing HgDiffParser with a diff that creates a new file"""
        diff = (
            b'diff -r bf544ea505f8 readme\n'
            b'--- /dev/null\n'
            b'+++ b/readme\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'readme',
            orig_file_details=PRE_CREATION,
            modified_filename=b'readme',
            modified_file_details=b'Uncommitted',
            data=diff)

    def test_diff_parser_with_added_empty_file(self) -> None:
        """Testing HgDiffParser with a diff with an added empty file"""
        diff = (
            b'diff -r 356a6127ef19 -r 4960455a8e88 empty\n'
            b'--- /dev/null\n'
            b'+++ b/empty\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'empty',
            orig_file_details=PRE_CREATION,
            modified_filename=b'empty',
            modified_file_details=b'4960455a8e88',
            data=diff)

    def test_diff_parser_with_deleted_empty_file(self) -> None:
        """Testing HgDiffParser with a diff with a deleted empty file"""
        diff = (
            b'diff -r 356a6127ef19 -r 4960455a8e88 empty\n'
            b'--- a/empty\n'
            b'+++ /dev/null\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'empty',
            orig_file_details=b'356a6127ef19',
            modified_filename=b'empty',
            modified_file_details=b'4960455a8e88',
            deleted=True,
            data=diff)

    def test_diff_parser_uncommitted(self) -> None:
        """Testing HgDiffParser with a diff with an uncommitted change"""
        diff = (
            b'diff -r bf544ea505f8 readme\n'
            b'--- a/readme\n'
            b'+++ b/readme\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'readme',
            orig_file_details=b'bf544ea505f8',
            modified_filename=b'readme',
            modified_file_details=b'Uncommitted',
            data=diff)

    def test_diff_parser_committed(self) -> None:
        """Testing HgDiffParser with a diff between committed revisions"""
        diff = (
            b'diff -r 356a6127ef19 -r 4960455a8e88 readme\n'
            b'--- a/readme\n'
            b'+++ b/readme\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'readme',
            orig_file_details=b'356a6127ef19',
            modified_filename=b'readme',
            modified_file_details=b'4960455a8e88',
            data=diff)

    def test_diff_parser_with_preamble_junk(self) -> None:
        """Testing HgDiffParser with a diff that contains non-diff junk text
        as a preamble
        """
        diff = (
            b'changeset:   60:3613c58ad1d5\n'
            b'user:        Michael Rowe <mrowe@mojain.com>\n'
            b'date:        Fri Jul 27 11:44:37 2007 +1000\n'
            b'files:       readme\n'
            b'description:\n'
            b'Update the readme file\n'
            b'\n'
            b'\n'
            b'diff -r 356a6127ef19 -r 4960455a8e88 readme\n'
            b'--- a/readme\n'
            b'+++ b/readme\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'readme',
            orig_file_details=b'356a6127ef19',
            modified_filename=b'readme',
            modified_file_details=b'4960455a8e88',
            data=diff)

    def test_git_diff_parsing(self) -> None:
        """Testing HgDiffParser git diff support"""
        diff = (
            b'# Node ID 4960455a8e88\n'
            b'# Parent bf544ea505f8\n'
            b'diff --git a/path/to file/readme.txt '
            b'b/new/path to/readme.txt\n'
            b'rename from path/to file/readme.txt\n'
            b'rename to new/path to/readme.txt\n'
            b'--- a/path/to file/readme.txt\n'
            b'+++ b/new/path to/readme.txt\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'path/to file/readme.txt',
            orig_file_details=b'bf544ea505f8',
            modified_filename=b'new/path to/readme.txt',
            modified_file_details=b'4960455a8e88',
            moved=True,
            data=diff)

    def test_git_diff_parser_with_new_binary_file(self) -> None:
        """Testing HgDiffParser diff with new binary file"""
        diff = (
            b'# Node ID 4960455a8e88\n'
            b'# Parent bf544ea505f8\n'
            b'diff --git a/test.png b/test.png\n'
            b'new file mode 100644\n'
            b'index 0000000000000000000000000000000000000000..'
            b'696eefb5e69b23fc2575f2eaf2863515f6937438\n'
            b'GIT binary patch\n'
            b'literal 12\n'
            b'zc$}1Zbx>SS@\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'test.png',
            orig_file_details=PRE_CREATION,
            modified_filename=b'test.png',
            modified_file_details=b'4960455a8e88',
            binary=True,
            new_unix_mode='100644',
            data=diff)

    def test_diff_parser_unicode(self) -> None:
        """Testing HgDiffParser with unicode characters"""
        diff = (
            'diff -r bf544ea505f8 réadme\n'
            '--- a/réadme\n'
            '+++ b/réadme\n'
        ).encode('utf-8')

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename='réadme'.encode('utf-8'),
            orig_file_details=b'bf544ea505f8',
            modified_filename='réadme'.encode('utf-8'),
            modified_file_details=b'Uncommitted',
            data=diff)

    def test_git_diff_parsing_unicode(self) -> None:
        """Testing HgDiffParser git diff with unicode characters"""
        diff = (
            '# Node ID 4960455a8e88\n'
            '# Parent bf544ea505f8\n'
            'diff --git a/path/to file/réadme.txt '
            'b/new/path to/réadme.txt\n'
            'rename from path/to file/réadme.txt\n'
            'rename to new/path to/réadme.txt\n'
            '--- a/path/to file/réadme.txt\n'
            '+++ b/new/path to/réadme.txt\n'
        ).encode('utf-8')

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename='path/to file/réadme.txt'.encode('utf-8'),
            orig_file_details=b'bf544ea505f8',
            modified_filename='new/path to/réadme.txt'.encode('utf-8'),
            modified_file_details=b'4960455a8e88',
            moved=True,
            data=diff)

    def test_revision_parsing(self) -> None:
        """Testing HgDiffParser revision number parsing"""
        self.assertEqual(
            self.tool.parse_diff_revision(filename=b'doc/readme',
                                          revision=b'bf544ea505f8'),
            (b'doc/readme', b'bf544ea505f8'))

        self.assertEqual(
            self.tool.parse_diff_revision(filename=b'/dev/null',
                                          revision=b'bf544ea505f8'),
            (b'/dev/null', PRE_CREATION))

    def test_get_branches(self) -> None:
        """Testing list of branches in HgClient.get_change"""
        value = self.tool.get_branches()
        self.assertTrue(isinstance(value, list))
        self.assertEqual(len(value), 1)

        self.assertEqual(value[0].id, 'default')
        self.assertEqual(value[0].commit,
                         '661e5dd3c4938ecbe8f77e2fdfa905d70485f94c')
        self.assertEqual(value[0].default, True)

    def test_get_change(self) -> None:
        """Testing raw diff of HgClient.get_change"""
        self.assertRaises(SCMError, lambda: self.tool.get_change('dummy'))

        value = self.tool.get_change('0')
        self.assertNotIn(b'goodbye', value.diff)
        self.assertEqual(value.id, 'f814b6e226d2ba6d26d02ca8edbff91f57ab2786')
        value = self.tool.get_change('1')
        self.assertIn(b'goodbye', value.diff)
        self.assertEqual(value.id, '661e5dd3c4938ecbe8f77e2fdfa905d70485f94c')

    def test_get_commits(self) -> None:
        """Testing commit objects in HgClient.get_commits"""
        value = self.tool.get_commits()
        self.assertTrue(isinstance(value, list))
        self.assertEqual(len(value), 2)

        self.assertEqual(value[0].id,
                         '661e5dd3c4938ecbe8f77e2fdfa905d70485f94c')
        self.assertEqual(value[0].message, 'second')
        self.assertEqual(value[0].author_name,
                         'Michael Rowe <mike.rowe@nab.com.au>')
        self.assertEqual(value[0].date, '2007-08-07T17:12:23')
        self.assertEqual(value[0].parent,
                         'f814b6e226d2ba6d26d02ca8edbff91f57ab2786')

        self.assertEqual(value[1].id,
                         'f814b6e226d2ba6d26d02ca8edbff91f57ab2786')
        self.assertEqual(value[1].message, 'first')
        self.assertEqual(value[1].author_name,
                         'Michael Rowe <mike.rowe@nab.com.au>')
        self.assertEqual(value[1].date, '2007-08-07T17:11:57')
        self.assertEqual(value[1].parent, '')

        self.assertRaisesRegex(SCMError, 'Cannot load commits: ',
                               lambda: self.tool.get_commits(branch='x'))

        rev = 'f814b6e226d2ba6d26d02ca8edbff91f57ab2786'
        value = self.tool.get_commits(start=rev)
        self.assertTrue(isinstance(value, list))
        self.assertEqual(len(value), 1)

    def test_get_commits_with_non_utc_server_timezone(self) -> None:
        """Testing commit objects in HgClient.get_commits with
        settings.TIME_ZONE != UTC
        """
        old_tz = os.environ['TZ']
        os.environ['TZ'] = 'US/Pacific'

        try:
            value = self.tool.get_commits()
        finally:
            os.environ['TZ'] = old_tz

        self.assertTrue(isinstance(value, list))
        self.assertEqual(len(value), 2)

        self.assertEqual(value[0].id,
                         '661e5dd3c4938ecbe8f77e2fdfa905d70485f94c')
        self.assertEqual(value[0].message, 'second')
        self.assertEqual(value[0].author_name,
                         'Michael Rowe <mike.rowe@nab.com.au>')
        self.assertEqual(value[0].date, '2007-08-07T17:12:23')
        self.assertEqual(value[0].parent,
                         'f814b6e226d2ba6d26d02ca8edbff91f57ab2786')

        self.assertEqual(value[1].id,
                         'f814b6e226d2ba6d26d02ca8edbff91f57ab2786')
        self.assertEqual(value[1].message, 'first')
        self.assertEqual(value[1].author_name,
                         'Michael Rowe <mike.rowe@nab.com.au>')
        self.assertEqual(value[1].date, '2007-08-07T17:11:57')
        self.assertEqual(value[1].parent, '')

        self.assertRaisesRegex(SCMError, 'Cannot load commits: ',
                               lambda: self.tool.get_commits(branch='x'))

        rev = 'f814b6e226d2ba6d26d02ca8edbff91f57ab2786'
        value = self.tool.get_commits(start=rev)
        self.assertTrue(isinstance(value, list))
        self.assertEqual(len(value), 1)

    def test_get_file(self) -> None:
        """Testing HgTool.get_file"""
        tool = self.tool

        value = tool.get_file('doc/readme', Revision('661e5dd3c493'))
        self.assertIsInstance(value, bytes)
        self.assertEqual(value, b'Hello\n\ngoodbye\n')

        with self.assertRaises(FileNotFoundError):
            tool.get_file('')

        with self.assertRaises(FileNotFoundError):
            tool.get_file('hello', PRE_CREATION)

    def test_file_exists(self) -> None:
        """Testing HgTool.file_exists"""
        rev = Revision('661e5dd3c493')

        self.assertTrue(self.tool.file_exists('doc/readme', rev))
        self.assertFalse(self.tool.file_exists('doc/readme2', rev))

    def test_get_file_base_commit_id_override(self) -> None:
        """Testing base_commit_id overrides revision in HgTool.get_file"""
        base_commit_id = '661e5dd3c493'
        bogus_rev = Revision('bogusrevision')
        file = 'doc/readme'

        context = FileLookupContext()
        context.base_commit_id = base_commit_id

        value = self.tool.get_file(file, bogus_rev, context=context)
        self.assertTrue(isinstance(value, bytes))
        self.assertEqual(value, b'Hello\n\ngoodbye\n')

        self.assertTrue(self.tool.file_exists(
            'doc/readme',
            bogus_rev,
            context=context))
        self.assertTrue(not self.tool.file_exists(
            'doc/readme2',
            bogus_rev,
            context=context))

    def test_interface(self) -> None:
        """Testing basic HgTool API"""
        self.assertTrue(self.tool.diffs_use_absolute_paths)

        self.assertRaises(NotImplementedError,
                          lambda: self.tool.get_changeset(1))

    @online_only
    def test_https_repo(self) -> None:
        """Testing HgTool.file_exists with an HTTPS-based repository"""
        repo = self.create_repository(
            name='Test HG2',
            path='https://hg-edge.mozilla.org/mozilla-central',
            tool_name='Mercurial')
        tool = repo.get_scmtool()

        self.assertTrue(tool.file_exists(
            'modules/libpref/init/StaticPrefList.yaml',
            Revision('e52ef55079f7'),
        ))
        self.assertFalse(tool.file_exists(
            'modules/libpref/init/StaticPrefList.yaml',
            Revision('abcdef123456'),
        ))

    def test_normalize_patch_with_git_diff_new_symlink(self) -> None:
        """Testing HgTool.normalize_patch with Git-style diff and new symlink
        """
        self.assertEqual(
            self.tool.normalize_patch(
                patch=(
                    b'diff --git /dev/null b/test\n'
                    b'new file mode 120000\n'
                    b'--- /dev/null\n'
                    b'+++ b/test\n'
                    b'@@ -0,0 +1,1 @@\n'
                    b'+target_file\n'
                    b'\\ No newline at end of file'
                ),
                filename='test',
                revision=PRE_CREATION),
            (
                b'diff --git /dev/null b/test\n'
                b'new file mode 100000\n'
                b'--- /dev/null\n'
                b'+++ b/test\n'
                b'@@ -0,0 +1,1 @@\n'
                b'+target_file\n'
                b'\\ No newline at end of file'
            ))

    def test_normalize_patch_with_git_diff_modified_symlink(self) -> None:
        """Testing HgTool.normalize_patch with Git-style diff and modified
        symlink
        """
        self.assertEqual(
            self.tool.normalize_patch(
                patch=(
                    b'diff --git a/test b/test\n'
                    b'index abc1234..def4567 120000\n'
                    b'--- a/test\n'
                    b'+++ b/test\n'
                    b'@@ -1,1 +1,1 @@\n'
                    b'-old_target\n'
                    b'\\ No newline at end of file'
                    b'+new_target\n'
                    b'\\ No newline at end of file'
                ),
                filename='test',
                revision='abc1234'),
            (
                b'diff --git a/test b/test\n'
                b'index abc1234..def4567 100000\n'
                b'--- a/test\n'
                b'+++ b/test\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-old_target\n'
                b'\\ No newline at end of file'
                b'+new_target\n'
                b'\\ No newline at end of file'
            ))

    def test_normalize_patch_with_git_diff_deleted_symlink(self) -> None:
        """Testing HgTool.normalize_patch with Git-style diff and deleted
        symlink
        """
        self.assertEqual(
            self.tool.normalize_patch(
                patch=(
                    b'diff --git a/test b/test\n'
                    b'deleted file mode 120000\n'
                    b'index abc1234..0000000\n'
                    b'--- a/test\n'
                    b'+++ /dev/null\n'
                    b'@@ -1,1 +0,0 @@\n'
                    b'-old_target\n'
                    b'\\ No newline at end of file'
                ),
                filename='test',
                revision='abc1234'),
            (
                b'diff --git a/test b/test\n'
                b'deleted file mode 100000\n'
                b'index abc1234..0000000\n'
                b'--- a/test\n'
                b'+++ /dev/null\n'
                b'@@ -1,1 +0,0 @@\n'
                b'-old_target\n'
                b'\\ No newline at end of file'
            ))

    def test_normalize_patch_with_hg_diff(self) -> None:
        """Testing HgTool.normalize_patch with Git-style diff and deleted
        symlink
        """
        self.assertEqual(
            self.tool.normalize_patch(
                patch=(
                    b'diff -r 123456789abc -r 123456789def test\n'
                    b'--- a/test\n'
                    b'+++ b/test\n'
                    b'@@ -1,1 +1,1 @@\n'
                    b'-a\n'
                    b'-b\n'
                ),
                filename='test',
                revision='123456789abc'),
            (
                b'diff -r 123456789abc -r 123456789def test\n'
                b'--- a/test\n'
                b'+++ b/test\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-a\n'
                b'-b\n'
            ))

    def test_get_diff_parser_cls_with_git_diff(self) -> None:
        """Testing HgTool._get_diff_parser_cls with Git diff"""
        self.assertIs(
            self.tool._get_diff_parser_cls(
                b'diff --git a/test b/test\n'
                b'--- a/test\n'
                b'+++ b/test\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-a\n'
                b'-b\n'),
            HgGitDiffParser)

    def test_get_diff_parser_cls_with_git_diff_and_header(self) -> None:
        """Testing HgTool._get_diff_parser_cls with Git diff and header"""
        self.assertIs(
            self.tool._get_diff_parser_cls(
                b'# HG changeset patch\n'
                b'# Node ID 123456789abc\n'
                b'# Parent cba987654321\n'
                b'diff --git a/test b/test\n'
                b'--- a/test\n'
                b'+++ b/test\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-a\n'
                b'-b\n'),
            HgGitDiffParser)

    def test_get_diff_parser_cls_with_hg_diff(self) -> None:
        """Testing HgTool._get_diff_parser_cls with Mercurial diff"""
        self.assertIs(
            self.tool._get_diff_parser_cls(
                b'diff -r 123456789abc -r 123456789def test\n'
                b'--- a/test   Thu Feb 17 12:36:00 2022 -0700\n'
                b'+++ b/test   Thu Feb 17 12:36:15 2022 -0700\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-a\n'
                b'-b\n'),
            HgDiffParser)

    def test_get_diff_parser_cls_with_hg_diff_and_header(self) -> None:
        """Testing HgTool._get_diff_parser_cls with Mercurial diff and header
        """
        self.assertIs(
            self.tool._get_diff_parser_cls(
                b'# HG changeset patch\n'
                b'# Node ID 123456789abc\n'
                b'# Parent cba987654321\n'
                b'diff -r 123456789abc -r 123456789def test\n'
                b'--- a/test   Thu Feb 17 12:36:00 2022 -0700\n'
                b'+++ b/test   Thu Feb 17 12:36:15 2022 -0700\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-a\n'
                b'-b\n'),
            HgDiffParser)

    def test_get_diff_parser_cls_with_git_before_hg(self) -> None:
        """Testing HgTool._get_diff_parser_cls with diff --git before diff -r
        """
        self.assertIs(
            self.tool._get_diff_parser_cls(
                b'diff --git a/test b/test\n'
                b'diff -r 123456789abc -r 123456789def test\n'),
            HgGitDiffParser)

    def test_get_diff_parser_cls_with_hg_before_git(self) -> None:
        """Testing HgTool._get_diff_parser_cls with diff -r before diff --git
        """
        self.assertIs(
            self.tool._get_diff_parser_cls(
                b'diff -r 123456789abc -r 123456789def test\n'
                b'diff --git a/test b/test\n'),
            HgDiffParser)


class HgWebClientTests(kgb.SpyAgency, TestCase):
    """Unit tests for reviewboard.scmtools.hg.HgWebClient."""

    def setUp(self) -> None:
        """Set up the test."""
        super().setUp()

        self.hgweb_client = HgWebClient(path='http://hg.example.com/',
                                        username='test-user',
                                        password='test-password')

    def test_cat_file_with_raw_file(self) -> None:
        """Testing HgWebClient.cat_file with URL using raw-file"""
        def _get_file_http(
            client: SCMClient,
            url: str,
            path: str,
            revision: RevisionID,
            mime_type: Optional[str],
            *args,
            **kwargs,
        ) -> Optional[bytes]:
            if url.startswith('http://hg.example.com/raw-file/'):
                return b'result payload'

            raise FileNotFoundError(path=path,
                                    revision=revision)

        spy = self.spy_on(self.hgweb_client.get_file_http,
                          call_fake=_get_file_http)

        rsp = self.hgweb_client.cat_file(path='foo/bar.txt',
                                         rev=HEAD)
        self.assertIsInstance(rsp, bytes)
        self.assertEqual(rsp, b'result payload')

        spy = self.hgweb_client.get_file_http.spy
        self.assertEqual(len(spy.calls), 1)
        self.assertTrue(spy.last_called_with(
            url='http://hg.example.com/raw-file/tip/foo/bar.txt',
            path='foo/bar.txt',
            revision='tip'))

    def test_cat_file_with_raw(self) -> None:
        """Testing HgWebClient.cat_file with URL using raw"""
        def _get_file_http(
            client: SCMClient,
            url: str,
            path: str,
            revision: RevisionID,
            mime_type: Optional[str],
            *args,
            **kwargs,
        ) -> Optional[bytes]:
            if url.startswith('http://hg.example.com/raw/'):
                return b'result payload'

            raise FileNotFoundError(path=path,
                                    revision=revision)

        spy = self.spy_on(self.hgweb_client.get_file_http,
                          call_fake=_get_file_http)

        rsp = self.hgweb_client.cat_file(path='foo/bar.txt',
                                         rev=HEAD)
        self.assertIsInstance(rsp, bytes)
        self.assertEqual(rsp, b'result payload')

        spy = self.hgweb_client.get_file_http.spy
        self.assertEqual(len(spy.calls), 2)
        self.assertTrue(spy.last_called_with(
            url='http://hg.example.com/raw/tip/foo/bar.txt',
            path='foo/bar.txt',
            revision='tip'))

    def test_cat_file_with_hg_history(self) -> None:
        """Testing HgWebClient.cat_file with URL using hg-history"""
        def _get_file_http(
            client: SCMClient,
            url: str,
            path: str,
            revision: RevisionID,
            mime_type: Optional[str],
            *args,
            **kwargs,
        ) -> Optional[bytes]:
            if url.startswith('http://hg.example.com/hg-history/'):
                return b'result payload'

            raise FileNotFoundError(path=path,
                                    revision=revision)

        self.spy_on(self.hgweb_client.get_file_http,
                    call_fake=_get_file_http)

        rsp = self.hgweb_client.cat_file(path='foo/bar.txt',
                                         rev=HEAD)
        self.assertIsInstance(rsp, bytes)
        self.assertEqual(rsp, b'result payload')

        spy = self.hgweb_client.get_file_http.spy
        self.assertEqual(len(spy.calls), 3)
        self.assertTrue(spy.last_called_with(
            url='http://hg.example.com/hg-history/tip/foo/bar.txt',
            path='foo/bar.txt',
            revision='tip'))

    def test_cat_file_with_base_commit_id(self) -> None:
        """Testing HgWebClient.cat_file with base_commit_id"""
        spy = self.spy_on(self.hgweb_client.get_file_http,
                          op=kgb.SpyOpReturn(b'result payload'))

        rsp = self.hgweb_client.cat_file(
            path='foo/bar.txt',
            base_commit_id='1ca5879492b8fd606df1964ea3c1e2f4520f076f')
        self.assertIsInstance(rsp, bytes)
        self.assertEqual(rsp, b'result payload')

        self.assertEqual(len(spy.calls), 1)
        self.assertTrue(spy.last_called_with(
            url='http://hg.example.com/raw-file/'
                '1ca5879492b8fd606df1964ea3c1e2f4520f076f/foo/bar.txt',
            path='foo/bar.txt',
            revision='1ca5879492b8fd606df1964ea3c1e2f4520f076f'))

    def test_cat_file_with_not_found(self) -> None:
        """Testing HgWebClient.cat_file with file not found"""
        spy = self.spy_on(
            self.hgweb_client.get_file_http,
            op=kgb.SpyOpRaise(FileNotFoundError('foo/bar.txt')))

        with self.assertRaises(FileNotFoundError):
            self.hgweb_client.cat_file(path='foo/bar.txt')

        self.assertEqual(len(spy.calls), 3)

    def test_get_branches(self) -> None:
        """Testing HgWebClient.get_branches"""
        def _get_file_http(
            client: SCMClient,
            url: str,
            path: str,
            revision: RevisionID,
            mime_type: Optional[str],
            *args,
            **kwargs,
        ) -> Optional[bytes]:
            self.assertEqual(url, 'http://hg.example.com/json-branches')
            self.assertEqual(mime_type, 'application/json')
            self.assertEqual(path, '')
            self.assertEqual(revision, '')

            return self._dump_json({
                'branches': [
                    {
                        'branch': 'default',
                        'node': '1ca5879492b8fd606df1964ea3c1e2f4520f076f',
                        'status': 'open',
                    },
                    {
                        'branch': 'closed-branch',
                        'node': 'b9af6489f6f2004ad11b82c6057f7007e3c35372',
                        'status': 'closed',
                    },
                    {
                        'branch': 'release-branch',
                        'node': '8210c0d945ef893d40a903c9dc14cd072eee5bb7',
                        'status': 'open',
                    },
                ],
            })

        self.spy_on(self.hgweb_client.get_file_http,
                    call_fake=_get_file_http)

        branches = self.hgweb_client.get_branches()
        self.assertIsInstance(branches, list)
        self.assertEqual(len(branches), 2)

        branch = branches[0]
        self.assertEqual(branch.id, 'default')
        self.assertEqual(branch.commit,
                         '1ca5879492b8fd606df1964ea3c1e2f4520f076f')
        self.assertTrue(branch.default)

        branch = branches[1]
        self.assertEqual(branch.id, 'release-branch')
        self.assertEqual(branch.commit,
                         '8210c0d945ef893d40a903c9dc14cd072eee5bb7')
        self.assertFalse(branch.default)

    def test_get_branches_with_error(self) -> None:
        """Testing HgWebClient.get_branches with error fetching result"""
        def _get_file_http(
            client: SCMClient,
            url: str,
            path: str,
            revision: RevisionID,
            mime_type: Optional[str],
            *args,
            **kwargs,
        ) -> Optional[bytes]:
            raise FileNotFoundError(path, revision)

        self.spy_on(self.hgweb_client.get_file_http,
                    call_fake=_get_file_http)

        branches = self.hgweb_client.get_branches()
        self.assertEqual(branches, [])

    def test_get_change(self) -> None:
        """Testing HgWebClient.get_change"""
        def _get_file_http(
            client: SCMClient,
            url: str,
            path: str,
            revision: RevisionID,
            mime_type: Optional[str],
            *args,
            **kwargs,
        ) -> Optional[bytes]:
            if url.startswith('http://hg.example.com/raw-rev/'):
                self.assertEqual(
                    url,
                    'http://hg.example.com/raw-rev/'
                    '1ca5879492b8fd606df1964ea3c1e2f4520f076f')
                self.assertEqual(path, '')
                self.assertEqual(revision, '')
                self.assertIsNone(mime_type)

                return b'diff payload'
            elif url.startswith('http://hg.example.com/json-rev/'):
                self.assertEqual(
                    url,
                    'http://hg.example.com/json-rev/'
                    '1ca5879492b8fd606df1964ea3c1e2f4520f076f')
                self.assertEqual(mime_type, 'application/json')
                self.assertEqual(path, '')
                self.assertEqual(revision, '')

                return self._dump_json({
                    'node': '1ca5879492b8fd606df1964ea3c1e2f4520f076f',
                    'desc': 'This is the change description',
                    'user': 'Test User',
                    'date': [1583149219, 28800],
                    'parents': [
                        'b9af6489f6f2004ad11b82c6057f7007e3c35372',
                    ],
                })
            else:
                raise FileNotFoundError(path=path,
                                        revision=revision)

        self.spy_on(self.hgweb_client.get_file_http,
                    call_fake=_get_file_http)

        commit = self.hgweb_client.get_change(
            '1ca5879492b8fd606df1964ea3c1e2f4520f076f')
        self.assertEqual(commit.id, '1ca5879492b8fd606df1964ea3c1e2f4520f076f')
        self.assertEqual(commit.message, 'This is the change description')
        self.assertEqual(commit.author_name, 'Test User')
        self.assertEqual(commit.date, '2020-03-02T03:40:19')
        self.assertEqual(commit.parent,
                         'b9af6489f6f2004ad11b82c6057f7007e3c35372')

    def test_get_commits(self) -> None:
        """Testing HgWebClient.get_commits"""
        def _get_file_http(
            client: SCMClient,
            url: str,
            path: str,
            revision: RevisionID,
            mime_type: Optional[str],
            *args,
            **kwargs,
        ) -> Optional[bytes]:
            self.assertEqual(
                url,
                'http://hg.example.com/json-log/?rev=branch(.)')
            self.assertEqual(mime_type, 'application/json')
            self.assertEqual(path, '')
            self.assertEqual(revision, '')

            return self._dump_json({
                'entries': [
                    {
                        'node': '1ca5879492b8fd606df1964ea3c1e2f4520f076f',
                        'desc': 'This is the change description',
                        'user': 'Test User',
                        'date': [1583149219, 28800],
                        'parents': [
                            'b9af6489f6f2004ad11b82c6057f7007e3c35372',
                        ],
                    },
                    {
                        'node': 'b9af6489f6f2004ad11b82c6057f7007e3c35372',
                        'desc': 'This is another description',
                        'user': 'Another User',
                        'date': [1581897120, 28800],
                        'parents': [
                            '8210c0d945ef893d40a903c9dc14cd072eee5bb7',
                        ],
                    },
                ],
            })

        self.spy_on(self.hgweb_client.get_file_http,
                    call_fake=_get_file_http)

        commits = self.hgweb_client.get_commits()
        self.assertEqual(len(commits), 2)

        commit = commits[0]
        self.assertEqual(commit.id, '1ca5879492b8fd606df1964ea3c1e2f4520f076f')
        self.assertEqual(commit.message, 'This is the change description')
        self.assertEqual(commit.author_name, 'Test User')
        self.assertEqual(commit.date, '2020-03-02T03:40:19')
        self.assertEqual(commit.parent,
                         'b9af6489f6f2004ad11b82c6057f7007e3c35372')

        commit = commits[1]
        self.assertEqual(commit.id, 'b9af6489f6f2004ad11b82c6057f7007e3c35372')
        self.assertEqual(commit.message, 'This is another description')
        self.assertEqual(commit.author_name, 'Another User')
        self.assertEqual(commit.date, '2020-02-16T15:52:00')
        self.assertEqual(commit.parent,
                         '8210c0d945ef893d40a903c9dc14cd072eee5bb7')

    def test_get_commits_with_branch(self) -> None:
        """Testing HgWebClient.get_commits with branch"""
        def _get_file_http(
            client: SCMClient,
            url: str,
            path: str,
            revision: RevisionID,
            mime_type: Optional[str],
            *args,
            **kwargs,
        ) -> Optional[bytes]:
            self.assertEqual(
                url,
                'http://hg.example.com/json-log/?rev=branch(my-branch)')
            self.assertEqual(mime_type, 'application/json')
            self.assertEqual(path, '')
            self.assertEqual(revision, '')

            return self._dump_json({
                'entries': [
                    {
                        'node': '1ca5879492b8fd606df1964ea3c1e2f4520f076f',
                        'desc': 'This is the change description',
                        'user': 'Test User',
                        'date': [1583149219, 28800],
                        'parents': [
                            'b9af6489f6f2004ad11b82c6057f7007e3c35372',
                        ],
                    },
                    {
                        'node': 'b9af6489f6f2004ad11b82c6057f7007e3c35372',
                        'desc': 'This is another description',
                        'user': 'Another User',
                        'date': [1581897120, 28800],
                        'parents': [
                            '8210c0d945ef893d40a903c9dc14cd072eee5bb7',
                        ],
                    },
                ],
            })

        self.spy_on(self.hgweb_client.get_file_http,
                    call_fake=_get_file_http)

        commits = self.hgweb_client.get_commits(branch='my-branch')
        self.assertEqual(len(commits), 2)

        commit = commits[0]
        self.assertEqual(commit.id, '1ca5879492b8fd606df1964ea3c1e2f4520f076f')
        self.assertEqual(commit.message, 'This is the change description')
        self.assertEqual(commit.author_name, 'Test User')
        self.assertEqual(commit.date, '2020-03-02T03:40:19')
        self.assertEqual(commit.parent,
                         'b9af6489f6f2004ad11b82c6057f7007e3c35372')

        commit = commits[1]
        self.assertEqual(commit.id, 'b9af6489f6f2004ad11b82c6057f7007e3c35372')
        self.assertEqual(commit.message, 'This is another description')
        self.assertEqual(commit.author_name, 'Another User')
        self.assertEqual(commit.date, '2020-02-16T15:52:00')
        self.assertEqual(commit.parent,
                         '8210c0d945ef893d40a903c9dc14cd072eee5bb7')

    def test_get_commits_with_start(self) -> None:
        """Testing HgWebClient.get_commits with start"""
        def _get_file_http(
            client: SCMClient,
            url: str,
            path: str,
            revision: RevisionID,
            mime_type: Optional[str],
            *args,
            **kwargs,
        ) -> Optional[bytes]:
            self.assertEqual(
                url,
                'http://hg.example.com/json-log/'
                '?rev=ancestors(1ca5879492b8fd606df1964ea3c1e2f4520f076f)'
                '+and+branch(.)')
            self.assertEqual(mime_type, 'application/json')
            self.assertEqual(path, '')
            self.assertEqual(revision, '')

            return self._dump_json({
                'entries': [
                    {
                        'node': '1ca5879492b8fd606df1964ea3c1e2f4520f076f',
                        'desc': 'This is the change description',
                        'user': 'Test User',
                        'date': [1583149219, 28800],
                        'parents': [
                            'b9af6489f6f2004ad11b82c6057f7007e3c35372',
                        ],
                    },
                    {
                        'node': 'b9af6489f6f2004ad11b82c6057f7007e3c35372',
                        'desc': 'This is another description',
                        'user': 'Another User',
                        'date': [1581897120, 28800],
                        'parents': [
                            '8210c0d945ef893d40a903c9dc14cd072eee5bb7',
                        ],
                    },
                ],
            })

        self.spy_on(self.hgweb_client.get_file_http,
                    call_fake=_get_file_http)

        commits = self.hgweb_client.get_commits(
            start='1ca5879492b8fd606df1964ea3c1e2f4520f076f')
        self.assertEqual(len(commits), 2)

        commit = commits[0]
        self.assertEqual(commit.id, '1ca5879492b8fd606df1964ea3c1e2f4520f076f')
        self.assertEqual(commit.message, 'This is the change description')
        self.assertEqual(commit.author_name, 'Test User')
        self.assertEqual(commit.date, '2020-03-02T03:40:19')
        self.assertEqual(commit.parent,
                         'b9af6489f6f2004ad11b82c6057f7007e3c35372')

        commit = commits[1]
        self.assertEqual(commit.id, 'b9af6489f6f2004ad11b82c6057f7007e3c35372')
        self.assertEqual(commit.message, 'This is another description')
        self.assertEqual(commit.author_name, 'Another User')
        self.assertEqual(commit.date, '2020-02-16T15:52:00')
        self.assertEqual(commit.parent,
                         '8210c0d945ef893d40a903c9dc14cd072eee5bb7')

    def test_get_commits_with_not_implemented(self) -> None:
        """Testing HgWebClient.get_commits with server response of "not yet
        implemented"
        """
        def _get_file_http(
            client: SCMClient,
            url: str,
            path: str,
            revision: RevisionID,
            mime_type: Optional[str],
            *args,
            **kwargs,
        ) -> Optional[bytes]:
            self.assertEqual(url,
                             'http://hg.example.com/json-log/?rev=branch(.)')
            self.assertEqual(mime_type, 'application/json')
            self.assertEqual(path, '')
            self.assertEqual(revision, '')

            return b'not yet implemented'

        self.spy_on(self.hgweb_client.get_file_http,
                    call_fake=_get_file_http)

        commits = self.hgweb_client.get_commits()
        self.assertEqual(commits, [])

    def _dump_json(
        self,
        obj: SerializableDjangoJSONDict,
    ) -> bytes:
        """Dump an object to a JSON byte string.

        Args:
            obj (dict):
                The object to dump.

        Returns:
            bytes;
            The JSON-serialized byte string.
        """
        return json.dumps(obj).encode('utf-8')


class HgAuthFormTests(TestCase):
    """Unit tests for HgTool's authentication form."""

    def test_fields(self) -> None:
        """Testing HgTool authentication form fields"""
        form = HgTool.create_auth_form()

        self.assertEqual(list(form.fields), ['username', 'password'])
        self.assertEqual(form['username'].help_text, '')
        self.assertEqual(form['username'].label, 'Username')
        self.assertEqual(form['password'].help_text, '')
        self.assertEqual(form['password'].label, 'Password')

    @add_fixtures(['test_scmtools'])
    def test_load(self) -> None:
        """Tetting HgTool authentication form load"""
        repository = self.create_repository(
            tool_name='Mercurial',
            username='test-user',
            password='test-pass')

        form = HgTool.create_auth_form(repository=repository)
        form.load()

        self.assertEqual(form['username'].value(), 'test-user')
        self.assertEqual(form['password'].value(), 'test-pass')

    @add_fixtures(['test_scmtools'])
    def test_save(self) -> None:
        """Tetting HgTool authentication form save"""
        repository = self.create_repository(tool_name='Mercurial')

        form = HgTool.create_auth_form(
            repository=repository,
            data={
                'username': 'test-user',
                'password': 'test-pass',
            })
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(repository.username, 'test-user')
        self.assertEqual(repository.password, 'test-pass')


class HgRepositoryFormTests(TestCase):
    """Unit tests for HgTool's repository form."""

    def test_fields(self) -> None:
        """Testing HgTool repository form fields"""
        form = HgTool.create_repository_form()

        self.assertEqual(list(form.fields), ['path', 'mirror_path'])
        self.assertEqual(form['path'].help_text,
                         'The path to the repository. This will generally be '
                         'the URL you would use to check out the repository.')
        self.assertEqual(form['path'].label, 'Path')
        self.assertEqual(form['mirror_path'].help_text, '')
        self.assertEqual(form['mirror_path'].label, 'Mirror Path')

    @add_fixtures(['test_scmtools'])
    def test_load(self) -> None:
        """Tetting HgTool repository form load"""
        repository = self.create_repository(
            tool_name='Mercurial',
            path='https://hg.example.com/repo',
            mirror_path='https://hg.mirror.example.com/repo')

        form = HgTool.create_repository_form(repository=repository)
        form.load()

        self.assertEqual(form['path'].value(), 'https://hg.example.com/repo')
        self.assertEqual(form['mirror_path'].value(),
                         'https://hg.mirror.example.com/repo')

    @add_fixtures(['test_scmtools'])
    def test_save(self) -> None:
        """Tetting HgTool repository form save"""
        repository = self.create_repository(tool_name='Mercurial')

        form = HgTool.create_repository_form(
            repository=repository,
            data={
                'path': 'https://hg.example.com/repo',
                'mirror_path': 'https://hg.mirror.example.com/repo',
            })
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(repository.path, 'https://hg.example.com/repo')
        self.assertEqual(repository.mirror_path,
                         'https://hg.mirror.example.com/repo')



"""Unit tests for reviewboard.scmtools.manager.RepositoryManager."""

from __future__ import annotations

from itertools import chain
from typing import Dict, List, Optional, Sequence, TYPE_CHECKING, Tuple, Union

from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import AnonymousUser, User
from djblets.testing.decorators import add_fixtures

from reviewboard.scmtools.models import Repository
from reviewboard.scmtools.testing.queries import (
    get_repositories_accessible_equeries,
    get_repositories_accessible_ids_equeries)
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from typelets.funcs import KwargsDict

    from reviewboard.site.models import AnyOrAllLocalSites

    _MixinParent = TestCase
else:
    _MixinParent = object


class AccessibleTestsMixin(_MixinParent):
    """Mixins for repository accessibility unit tests.

    Version Added:
        5.0.7
    """

    def _create_accessible_repository_data(
        self,
        *,
        user: Union[AnonymousUser, User],
        with_local_sites: bool = False,
        with_member: bool = False,
        with_member_by_group: bool = False,
        group_kwargs: KwargsDict = {},
    ) -> Tuple[Sequence[Repository], Optional[LocalSite]]:
        """Create test repository data for accessibility checks.

        This will create repositories on the global site and, optionally,
        two Local Sites. The provided user may optionally be granted membership
        directly or via group.

        Args:
            user (django.contrib.auth.models.AnonymousUser or
                  django.contrib.auth.models.User):
                The user to associate with any membership lists.

            with_local_sites (bool, optional):
                Whether to create Local Sites.

            with_member (bool, optional):
                Whether to make the user a member of the repository.

            with_member_by_group (bool, optional):
                Whether to make the user a member of the repository via a
                review group.

            group_kwargs (dict, optional):
                Additional keyword arguments to pass to review group
                construction.

        Returns:
            tuple:
            A 2-tuple of:

            Tuple:
                0 (list of reviewboard.scmtools.models.Repository):
                    The list of created repositories, in order.

                1 (reviewboard.site.models.LocalSite):
                    The first Local Site created, or ``None`` if not creating
                    Local Sites.
        """
        repositories_by_site: Dict[Optional[LocalSite], List[Repository]] = {}
        local_sites: List[Optional[LocalSite]] = []
        local_site_kwargs: KwargsDict = {}
        repository_i = 1

        if user.is_authenticated:
            # Satisfy the type checker.
            assert isinstance(user, User)

            local_site_kwargs['users'] = [user]

            if with_member_by_group:
                group_kwargs['users'] = [user]

        if with_local_sites:
            # Create the Local Sites
            local_sites += [
                self.create_local_site(name='test-site-1',
                                       **local_site_kwargs),
                self.create_local_site(name='test-site-2',
                                       **local_site_kwargs),
            ]

        # Add the global site last.
        local_sites.append(None)

        for local_site in local_sites:
            repositories_by_site[local_site] = [
                self.create_repository(local_site=local_site,
                                       name=f'repo{repository_i}',
                                       public=True,
                                       visible=True),
                self.create_repository(local_site=local_site,
                                       name=f'repo{repository_i + 1}',
                                       public=True,
                                       visible=False),
                self.create_repository(local_site=local_site,
                                       name=f'repo{repository_i + 2}',
                                       public=False,
                                       visible=True),
                self.create_repository(local_site=local_site,
                                       name=f'repo{repository_i + 3}',
                                       public=False,
                                       visible=False),
            ]
            repository_i += 4

        if with_member and user.is_authenticated:
            # Satisfy the type checker.
            assert isinstance(user, User)

            for local_site in local_sites:
                user.repositories.add(*repositories_by_site[local_site])

        if with_member_by_group:
            for local_site in local_sites:
                group = self.create_review_group(local_site=local_site,
                                                 **group_kwargs)
                group.repositories.add(*repositories_by_site[local_site])

        return (
            list(chain.from_iterable(repositories_by_site.values())),
            local_sites[0],
        )


class AccessibleTests(AccessibleTestsMixin, TestCase):
    """Unit tests for RepositoryManager.accessible()."""

    fixtures = ['test_scmtools', 'test_users']

    #
    # Anonymous tests
    #

    def test_with_anonymous(self) -> None:
        """Testing Repository.objects.accessible with anonymous and
        visible_only=False
        """
        user = AnonymousUser()
        repositories = self._create_accessible_repository_data(user=user)[0]

        self._test_accessible(
            repositories=[
                repositories[0],
                repositories[1],
            ],
            user=user,
            visible_only=False)

    def test_with_anonymous_visible_only(self) -> None:
        """Testing Repository.objects.accessible with anonymous and
        visible_only=True
        """
        user = AnonymousUser()
        repositories = self._create_accessible_repository_data(user=user)[0]

        self._test_accessible(
            repositories=[repositories[0]],
            user=user,
            visible_only=True)

    #
    # Superuser tests
    #

    def test_with_superuser(self) -> None:
        """Testing Repository.objects.accessible with superuser and
        visible_only=False
        """
        user = self.create_user(is_superuser=True)
        repositories = self._create_accessible_repository_data(user=user)[0]

        self._test_accessible(
            repositories=repositories,
            user=user,
            visible_only=False)

    def test_with_superuser_visible_only(self) -> None:
        """Testing Repository.objects.accessible with superuser and
        visible_only=True
        """
        user = self.create_user(is_superuser=True)
        repositories = self._create_accessible_repository_data(user=user)[0]

        # Note that, unlike with standard members, superusers won't see
        # hidden repositories in this case. This is a legacy
        # piece of logic in Group.objects.accessible().
        self._test_accessible(
            repositories=[
                repositories[0],
                repositories[2],
            ],
            user=user,
            visible_only=True)

    def test_with_superuser_local_site(self) -> None:
        """Testing Repository.objects.accessible with superuser,
        Local Site, and visible_only=False
        """
        user = self.create_user(is_superuser=True)
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)

        self._test_accessible(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=False)

    def test_with_superuser_local_site_visible_only(self) -> None:
        """Testing Repository.objects.accessible with superuser and
        visible_only=True
        """
        user = self.create_user(is_superuser=True)
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)

        # Note that, unlike with standard members, superusers won't see
        # hidden repositories in this case. This is a legacy
        # piece of logic in Group.objects.accessible().
        self._test_accessible(
            repositories=[
                repositories[0],
                repositories[2],
            ],
            user=user,
            local_site=local_site,
            visible_only=True)

    def test_with_superuser_local_site_all(self) -> None:
        """Testing Repository.objects.accessible with superuser,
        all Local Sites, and visible_only=False
        """
        user = self.create_user(is_superuser=True)
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)[0]

        self._test_accessible(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    def test_with_superuser_local_site_all_visible_only(self) -> None:
        """Testing Repository.objects.accessible with superuser, all
        Local Sites, and visible_only=True
        """
        user = self.create_user(is_superuser=True)
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)[0]

        # Note that, unlike with standard members, superusers won't see
        # hidden repositories in this case. This is a legacy
        # piece of logic in Group.objects.accessible().
        self._test_accessible(
            repositories=[
                repositories[0],
                repositories[2],
                repositories[4],
                repositories[6],
                repositories[8],
                repositories[10],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    #
    # Non-member tests
    #

    def test_with_non_member(self) -> None:
        """Testing Repository.objects.accessible with non-member and
        visible_only=False
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(user=user)[0]

        self._test_accessible(
            repositories=[
                repositories[0],
                repositories[1],
            ],
            user=user,
            visible_only=False)

    def test_with_non_member_visible_only(self) -> None:
        """Testing Repository.objects.accessible with non-member and
        visible_only=True
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(user=user)[0]

        self._test_accessible(
            repositories=[repositories[0]],
            user=user,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_non_member_local_site(self) -> None:
        """Testing Repository.objects.accessible with non-member,
        Local Site, and visible_only=False
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)

        self._test_accessible(
            repositories=[
                repositories[0],
                repositories[1],
            ],
            user=user,
            local_site=local_site,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_non_member_local_site_visible_only(self) -> None:
        """Testing Repository.objects.accessible with non-member,
        Local Site, and visible_only=True
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)

        self._test_accessible(
            repositories=[repositories[0]],
            user=user,
            local_site=local_site,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_non_member_local_site_all(self) -> None:
        """Testing Repository.objects.accessible with non-member,
        all Local Sites, and visible_only=False
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)

        self._test_accessible(
            repositories=[
                repositories[0],
                repositories[1],
                repositories[4],
                repositories[5],
                repositories[8],
                repositories[9],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_non_member_local_site_all_visible_only(self) -> None:
        """Testing Repository.objects.accessible with non-member,
        all Local Sites, and visible_only=True
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)

        self._test_accessible(
            repositories=[
                repositories[0],
                repositories[4],
                repositories[8],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    #
    # Member tests
    #

    def test_with_member(self) -> None:
        """Testing Repository.objects.accessible with member,
        visible_only=False
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_member=True)[0]

        self._test_accessible(
            repositories=repositories[:4],
            user=user,
            visible_only=False)

    def test_with_member_visible_only(self) -> None:
        """Testing Repository.objects.accessible with member,
        visible_only=True
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_member=True)[0]

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible().
        self._test_accessible(
            repositories=repositories[:4],
            user=user,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_member_local_site(self) -> None:
        """Testing Repository.objects.accessible with member, Local Site,
        and visible_only=False
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member=True)

        self._test_accessible(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_member_local_site_visible_only(self) -> None:
        """Testing Repository.objects.accessible with member, Local Site,
        and visible_only=True
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member=True)

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible().
        self._test_accessible(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_member_local_site_all(self) -> None:
        """Testing Repository.objects.accessible with member, all
        Local Sites, and visible_only=False
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member=True)

        self._test_accessible(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_member_local_site_all_visible_only(self) -> None:
        """Testing Repository.objects.accessible with member, all
        Local Sites, and visible_only=True
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member=True)

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible().
        self._test_accessible(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    #
    # Member by public group tests
    #

    def test_with_member_by_group(self) -> None:
        """Testing Repository.objects.accessible with member by group,
        and visible_only=False
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_member_by_group=True)[0]

        self._test_accessible(
            repositories=repositories[:4],
            user=user,
            visible_only=False)

    def test_with_member_by_group_visible_only(self) -> None:
        """Testing Repository.objects.accessible with member by group,
        and visible_only=True
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_member_by_group=True)[0]

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible().
        self._test_accessible(
            repositories=repositories[:4],
            user=user,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_member_by_group_local_site(self) -> None:
        """Testing Repository.objects.accessible with member by group,
        Local Site, and visible_only=False
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True)

        self._test_accessible(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_member_by_group_local_site_visible_only(self) -> None:
        """Testing Repository.objects.accessible with member by group,
        Local Site, and visible_only=True
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True)

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible().
        self._test_accessible(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_member_by_group_local_site_all(self) -> None:
        """Testing Repository.objects.accessible with member by group,
        all Local Sites, visible_only=False
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True)[0]

        self._test_accessible(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_member_by_group_local_site_all_visible_only(
        self,
    ) -> None:
        """Testing Repository.objects.accessible with member by group,
        all Local Sites, visible_only=True
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True)[0]

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible().
        self._test_accessible(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    #
    # Member by invite-only group tests
    #

    def test_with_member_by_invite_only_group(self) -> None:
        """Testing Repository.objects.accessible with member by group,
        and visible_only=False
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_member_by_group=True,
            group_kwargs={
                'invite_only': True,
            })[0]

        self._test_accessible(
            repositories=repositories[:4],
            user=user,
            visible_only=False)

    def test_with_member_by_invite_only_group_visible_only(self) -> None:
        """Testing Repository.objects.accessible with member by group,
        and visible_only=True
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_member_by_group=True,
            group_kwargs={
                'invite_only': True,
            })[0]

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible().
        self._test_accessible(
            repositories=repositories[:4],
            user=user,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_member_by_invite_only_group_local_site(self) -> None:
        """Testing Repository.objects.accessible with member by group,
        Local Site, and visible_only=False
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True,
            group_kwargs={
                'invite_only': True,
            })

        self._test_accessible(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_member_by_invite_only_group_local_site_visible_only(
        self,
    ) -> None:
        """Testing Repository.objects.accessible with member by group,
        Local Site, and visible_only=True
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True,
            group_kwargs={
                'invite_only': True,
            })

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible().
        self._test_accessible(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_member_by_invite_only_group_local_site_all(self) -> None:
        """Testing Repository.objects.accessible with member by group,
        all Local Sites, visible_only=False
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True,
            group_kwargs={
                'invite_only': True,
            })[0]

        self._test_accessible(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_member_by_invite_only_group_local_site_all_visible_only(
        self,
    ) -> None:
        """Testing Repository.objects.accessible with member by group,
        all Local Sites, visible_only=True
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True,
            group_kwargs={
                'invite_only': True,
            })[0]

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible().
        self._test_accessible(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    @add_fixtures(['test_users', 'test_site'])
    def test_with_show_all_local_sites(self) -> None:
        """Testing Group.objects.accessible with show_all_local_sites=True
        """
        user = self.create_user(is_superuser=True)

        repository = self.create_repository(with_local_site=True)
        repository.local_site.users.add(user)

        self.assertIn(
            repository,
            Repository.objects.accessible(user,
                                          local_site=repository.local_site))
        self.assertIn(
            repository,
            Repository.objects.accessible(user,
                                          local_site=LocalSite.ALL))

    def _test_accessible(
        self,
        *,
        repositories: Sequence[Repository],
        user: Union[AnonymousUser, User],
        visible_only: bool,
        local_site: AnyOrAllLocalSites = None,
    ) -> None:
        """Test the results of accessible_ids(), using the given settings.

        Version Added:
            5.0.7

        Args:
            repository (list of reviewboard.scmtools.models.Repository):
                The repositories to expect in results.

            user (django.contrib.auth.models.AnonymousUser or
                  django.contrib.auth.models.User):
                The user to check for access.

            visible_only (bool):
                Whether to check for visible repositories only.

            local_site (reviewboard.site.models.LocalSite, optional):
                The optional Local Site to check.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        # Prime the caches.
        if user.is_authenticated:
            user.get_local_site_stats()
            user.get_profile()

            if local_site is not LocalSite.ALL:
                user.get_site_profile(local_site=local_site)

        equeries = get_repositories_accessible_equeries(
            user=user,
            visible_only=visible_only,
            local_site=local_site)

        with assert_queries(equeries):
            accessible = list(Repository.objects.accessible(
                user,
                visible_only=visible_only,
                local_site=local_site))

        self.assertEqual(accessible, repositories)


class AccessibleIDsTests(AccessibleTestsMixin, TestCase):
    """Unit tests for RepositoryManager.accessible_ids()."""

    fixtures = ['test_scmtools', 'test_users']

    #
    # Anonymous tests
    #

    def test_with_anonymous(self) -> None:
        """Testing Repository.objects.accessible_ids with anonymous and
        visible_only=False
        """
        user = AnonymousUser()
        repositories = self._create_accessible_repository_data(user=user)[0]

        self._test_accessible_ids(
            repositories=[
                repositories[0],
                repositories[1],
            ],
            user=user,
            visible_only=False)

    def test_with_anonymous_visible_only(self) -> None:
        """Testing Repository.objects.accessible_ids with anonymous and
        visible_only=True
        """
        user = AnonymousUser()
        repositories = self._create_accessible_repository_data(user=user)[0]

        self._test_accessible_ids(
            repositories=[repositories[0]],
            user=user,
            visible_only=True)

    #
    # Superuser tests
    #

    def test_with_superuser(self) -> None:
        """Testing Repository.objects.accessible_ids with superuser and
        visible_only=False
        """
        user = self.create_user(is_superuser=True)
        repositories = self._create_accessible_repository_data(user=user)[0]

        self._test_accessible_ids(
            repositories=repositories,
            user=user,
            visible_only=False)

    def test_with_superuser_visible_only(self) -> None:
        """Testing Repository.objects.accessible_ids with superuser and
        visible_only=True
        """
        user = self.create_user(is_superuser=True)
        repositories = self._create_accessible_repository_data(user=user)[0]

        # Note that, unlike with standard members, superusers won't see
        # hidden repositories in this case. This is a legacy
        # piece of logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            repositories=[
                repositories[0],
                repositories[2],
            ],
            user=user,
            visible_only=True)

    def test_with_superuser_local_site(self) -> None:
        """Testing Repository.objects.accessible_ids with superuser,
        Local Site, and visible_only=False
        """
        user = self.create_user(is_superuser=True)
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)

        self._test_accessible_ids(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=False)

    def test_with_superuser_local_site_visible_only(self) -> None:
        """Testing Repository.objects.accessible_ids with superuser and
        visible_only=True
        """
        user = self.create_user(is_superuser=True)
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)

        # Note that, unlike with standard members, superusers won't see
        # hidden repositories in this case. This is a legacy
        # piece of logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            repositories=[
                repositories[0],
                repositories[2],
            ],
            user=user,
            local_site=local_site,
            visible_only=True)

    def test_with_superuser_local_site_all(self) -> None:
        """Testing Repository.objects.accessible_ids with superuser,
        all Local Sites, and visible_only=False
        """
        user = self.create_user(is_superuser=True)
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)[0]

        self._test_accessible_ids(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    def test_with_superuser_local_site_all_visible_only(self) -> None:
        """Testing Repository.objects.accessible_ids with superuser,
        all Local Sites, and visible_only=True
        """
        user = self.create_user(is_superuser=True)
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)[0]

        # Note that, unlike with standard members, superusers won't see
        # hidden repositories in this case. This is a legacy
        # piece of logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            repositories=[
                repositories[0],
                repositories[2],
                repositories[4],
                repositories[6],
                repositories[8],
                repositories[10],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    #
    # Non-member tests
    #

    def test_with_non_member(self) -> None:
        """Testing Repository.objects.accessible_ids with non-member and
        visible_only=False
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(user=user)[0]

        self._test_accessible_ids(
            repositories=[
                repositories[0],
                repositories[1],
            ],
            user=user,
            visible_only=False)

    def test_with_non_member_visible_only(self) -> None:
        """Testing Repository.objects.accessible_ids with non-member and
        visible_only=True
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(user=user)[0]

        self._test_accessible_ids(
            repositories=[repositories[0]],
            user=user,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_non_member_local_site(self) -> None:
        """Testing Repository.objects.accessible_ids with non-member,
        Local Site, and visible_only=False
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)

        self._test_accessible_ids(
            repositories=[
                repositories[0],
                repositories[1],
            ],
            user=user,
            local_site=local_site,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_non_member_local_site_visible_only(self) -> None:
        """Testing Repository.objects.accessible_ids with non-member,
        Local Site, and visible_only=True
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)

        self._test_accessible_ids(
            repositories=[repositories[0]],
            user=user,
            local_site=local_site,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_non_member_local_site_all(self) -> None:
        """Testing Repository.objects.accessible_ids with non-member,
        all Local Sites, and visible_only=False
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)

        self._test_accessible_ids(
            repositories=[
                repositories[0],
                repositories[1],
                repositories[4],
                repositories[5],
                repositories[8],
                repositories[9],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_non_member_local_site_all_visible_only(self) -> None:
        """Testing Repository.objects.accessible_ids with non-member,
        all Local Sites, and visible_only=True
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True)

        self._test_accessible_ids(
            repositories=[
                repositories[0],
                repositories[4],
                repositories[8],
            ],
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    #
    # Member tests
    #

    def test_with_member(self) -> None:
        """Testing Repository.objects.accessible_ids with member,
        visible_only=False
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_member=True)[0]

        self._test_accessible_ids(
            repositories=repositories,
            user=user,
            visible_only=False)

    def test_with_member_visible_only(self) -> None:
        """Testing Repository.objects.accessible_ids with member,
        visible_only=True
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_member=True)[0]

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            repositories=repositories,
            user=user,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_member_local_site(self) -> None:
        """Testing Repository.objects.accessible_ids with member, Local Site,
        and visible_only=False
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member=True)

        self._test_accessible_ids(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_member_local_site_visible_only(self) -> None:
        """Testing Repository.objects.accessible_ids with member, Local Site,
        and visible_only=True
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member=True)

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_member_local_site_all(self) -> None:
        """Testing Repository.objects.accessible_ids with member, all
        Local Sites, and visible_only=False
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member=True)[0]

        self._test_accessible_ids(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_member_local_site_all_visible_only(self) -> None:
        """Testing Repository.objects.accessible_ids with member, all
        Local Sites, and visible_only=True
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member=True)[0]

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    #
    # Member by public group tests
    #

    def test_with_member_by_group(self) -> None:
        """Testing Repository.objects.accessible_ids with member by group,
        and visible_only=False
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_member_by_group=True)[0]

        self._test_accessible_ids(
            repositories=repositories[:4],
            user=user,
            visible_only=False)

    def test_with_member_by_group_visible_only(self) -> None:
        """Testing Repository.objects.accessible_ids with member by group,
        and visible_only=True
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_member_by_group=True)[0]

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            repositories=repositories[:4],
            user=user,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_member_by_group_local_site(self) -> None:
        """Testing Repository.objects.accessible_ids with member by group,
        Local Site, and visible_only=False
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True)

        self._test_accessible_ids(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_member_by_group_local_site_visible_only(self) -> None:
        """Testing Repository.objects.accessible_ids with member by group,
        Local Site, and visible_only=True
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True)

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_member_by_group_local_site_all(self) -> None:
        """Testing Repository.objects.accessible_ids with member by group,
        all Local Sites, visible_only=False
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True)[0]

        self._test_accessible_ids(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_member_by_group_local_site_all_visible_only(
        self,
    ) -> None:
        """Testing Repository.objects.accessible_ids with member by group,
        all Local Sites, visible_only=True
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True)[0]

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    #
    # Member by invite-only group tests
    #

    def test_with_member_by_invite_only_group(self) -> None:
        """Testing Repository.objects.accessible_ids with member by group,
        and visible_only=False
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_member_by_group=True,
            group_kwargs={
                'invite_only': True,
            })[0]

        self._test_accessible_ids(
            repositories=repositories[:4],
            user=user,
            visible_only=False)

    def test_with_member_by_invite_only_group_visible_only(self) -> None:
        """Testing Repository.objects.accessible_ids with member by group,
        and visible_only=True
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_member_by_group=True,
            group_kwargs={
                'invite_only': True,
            })[0]

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            repositories=repositories[:4],
            user=user,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_member_by_invite_only_group_local_site(self) -> None:
        """Testing Repository.objects.accessible_ids with member by group,
        Local Site, and visible_only=False
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True,
            group_kwargs={
                'invite_only': True,
            })

        self._test_accessible_ids(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_member_by_invite_only_group_local_site_visible_only(
        self,
    ) -> None:
        """Testing Repository.objects.accessible_ids with member by group,
        Local Site, and visible_only=True
        """
        user = self.create_user()
        repositories, local_site = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True,
            group_kwargs={
                'invite_only': True,
            })

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            repositories=repositories[:4],
            user=user,
            local_site=local_site,
            visible_only=True)

    @add_fixtures(['test_site'])
    def test_with_member_by_invite_only_group_local_site_all(self) -> None:
        """Testing Repository.objects.accessible_ids with member by group,
        all Local Sites, visible_only=False
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True,
            group_kwargs={
                'invite_only': True,
            })[0]

        self._test_accessible_ids(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=False)

    @add_fixtures(['test_site'])
    def test_with_member_by_invite_only_group_local_site_all_visible_only(
        self,
    ) -> None:
        """Testing Repository.objects.accessible_ids with member by group,
        all Local Sites, visible_only=True
        """
        user = self.create_user()
        repositories = self._create_accessible_repository_data(
            user=user,
            with_local_sites=True,
            with_member_by_group=True,
            group_kwargs={
                'invite_only': True,
            })[0]

        # Note that members can see hidden repositories. This is a legacy
        # piece of intentional logic in Group.objects.accessible_ids().
        self._test_accessible_ids(
            repositories=repositories,
            user=user,
            local_site=LocalSite.ALL,
            visible_only=True)

    @add_fixtures(['test_users', 'test_site'])
    def test_with_show_all_local_sites(self) -> None:
        """Testing Group.objects.accessible_ids with show_all_local_sites=True
        """
        user = self.create_user(is_superuser=True)

        repository = self.create_repository(with_local_site=True)
        repository.local_site.users.add(user)

        self.assertIn(
            repository.pk,
            Repository.objects.accessible_ids(
                user,
                local_site=repository.local_site))
        self.assertIn(
            repository.pk,
            Repository.objects.accessible_ids(
                user,
                local_site=LocalSite.ALL))

    def _test_accessible_ids(
        self,
        *,
        repositories: Sequence[Repository],
        user: Union[AnonymousUser, User],
        visible_only: bool,
        local_site: AnyOrAllLocalSites = None,
        expect_in: bool = True,
    ) -> None:
        """Test the results of accessible_ids(), using the given settings.

        Version Added:
            5.0.7

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repositories to expect in results.

            user (django.contrib.auth.models.AnonymousUser or
                  django.contrib.auth.models.User):
                The user to check for access.

            visible_only (bool):
                Whether to check for visible repositories only.

            local_site (reviewboard.site.models.LocalSite, optional):
                The optional Local Site to check.

            expect_in (bool, optional):
                Whether to expect the provided repository's ID to be in the
                accessible IDs list.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        # Prime the caches.
        if user.is_authenticated:
            user.get_local_site_stats()
            user.get_profile()

            if local_site is not LocalSite.ALL:
                user.get_site_profile(local_site=local_site)

        equeries = get_repositories_accessible_ids_equeries(
            user=user,
            visible_only=visible_only,
            local_site=local_site)

        with assert_queries(equeries):
            accessible_ids = Repository.objects.accessible_ids(
                user,
                visible_only=visible_only,
                local_site=local_site)

        self.assertEqual(
            accessible_ids,
            [
                repository.pk
                for repository in repositories
            ])


class GetBestMatchTests(TestCase):
    """Unit tests for RepositoryManager.get_best_match()."""

    fixtures = ['test_scmtools', 'test_users']

    def test_with_pk(self):
        """Testing Repository.objects.get_best_match with repository ID"""
        repository1 = self.create_repository()
        self.create_repository(name=str(repository1.pk))

        self.assertEqual(
            Repository.objects.get_best_match(repository1.pk),
            repository1)

    @add_fixtures(['test_site'])
    def test_with_pk_and_local_site(self):
        """Testing Repository.objects.get_best_match with repository ID and
        local_site=...
        """
        repository1 = self.create_repository(with_local_site=True)
        repository2 = self.create_repository()
        local_site = repository1.local_site

        # This should match.
        self.assertEqual(
            Repository.objects.get_best_match(repository1.pk,
                                              local_site=local_site),
            repository1)

        # These should both fail.
        with self.assertRaises(Repository.DoesNotExist):
            Repository.objects.get_best_match(repository1.pk)

        with self.assertRaises(Repository.DoesNotExist):
            Repository.objects.get_best_match(repository2.pk,
                                              local_site=local_site)

    def test_with_name(self):
        """Testing Repository.objects.get_best_match with repository name"""
        repository1 = self.create_repository(name='repo 1')
        self.create_repository(name='repo 2')

        self.assertEqual(
            Repository.objects.get_best_match('repo 1'),
            repository1)

    @add_fixtures(['test_site'])
    def test_with_name_and_local_site(self):
        """Testing Repository.objects.get_best_match with repository name
        and local_site=...
        """
        repository1 = self.create_repository(name='repo 1',
                                             with_local_site=True)
        repository2 = self.create_repository(name='repo 2')
        local_site = repository1.local_site

        # This should match.
        self.assertEqual(
            Repository.objects.get_best_match(repository1.name,
                                              local_site=local_site),
            repository1)

        # These should both fail.
        with self.assertRaises(Repository.DoesNotExist):
            Repository.objects.get_best_match(repository1.name)

        with self.assertRaises(Repository.DoesNotExist):
            Repository.objects.get_best_match(repository2.name,
                                              local_site=local_site)

    def test_with_path(self):
        """Testing Repository.objects.get_best_match with repository path"""
        repository1 = self.create_repository(name='repo1',
                                             path='/test-path-1')
        self.create_repository(name='repo2',
                               path='/test-path-2')

        self.assertEqual(
            Repository.objects.get_best_match('/test-path-1'),
            repository1)

    @add_fixtures(['test_site'])
    def test_with_path_and_local_site(self):
        """Testing Repository.objects.get_best_match with repository path
        and local_site=...
        """
        repository1 = self.create_repository(path='/test-path-1',
                                             with_local_site=True)
        repository2 = self.create_repository(path='/test-path-2')
        local_site = repository1.local_site

        # This should match.
        self.assertEqual(
            Repository.objects.get_best_match(repository1.path,
                                              local_site=local_site),
            repository1)

        # These should both fail.
        with self.assertRaises(Repository.DoesNotExist):
            Repository.objects.get_best_match(repository1.path)

        with self.assertRaises(Repository.DoesNotExist):
            Repository.objects.get_best_match(repository2.path,
                                              local_site=local_site)

    def test_with_mirror_path(self):
        """Testing Repository.objects.get_best_match with repository
        mirror_path
        """
        repository1 = self.create_repository(name='repo1',
                                             mirror_path='/test-path-1')
        self.create_repository(name='repo2',
                               mirror_path='/test-path-2')

        self.assertEqual(
            Repository.objects.get_best_match('/test-path-1'),
            repository1)

    @add_fixtures(['test_site'])
    def test_with_mirror_path_and_local_site(self):
        """Testing Repository.objects.get_best_match with repository
        mirror_path and local_site=...
        """
        repository1 = self.create_repository(name='repo1',
                                             mirror_path='/test-path-1',
                                             with_local_site=True)
        repository2 = self.create_repository(name='repo2',
                                             mirror_path='/test-path-2')
        local_site = repository1.local_site

        # This should match.
        self.assertEqual(
            Repository.objects.get_best_match(repository1.mirror_path,
                                              local_site=local_site),
            repository1)

        # These should both fail.
        with self.assertRaises(Repository.DoesNotExist):
            Repository.objects.get_best_match(repository1.mirror_path)

        with self.assertRaises(Repository.DoesNotExist):
            Repository.objects.get_best_match(repository2.mirror_path,
                                              local_site=local_site)

    def test_with_no_match(self):
        """Testing Repository.objects.get_best_match with no match"""
        self.create_repository(name='repo 1')
        self.create_repository(name='repo 2')

        with self.assertRaises(Repository.DoesNotExist):
            Repository.objects.get_best_match('bad-id')

    def test_with_multiple_prefer_visible(self):
        """Testing Repository.objects.get_best_match with multiple results
        prefers visible over name/path/mirror_path
        """
        repository1 = self.create_repository(
            name='repo1',
            path='/path1',
            mirror_path='mirror')
        repository2 = self.create_repository(
            name='repo2',
            path='/path2',
            mirror_path='mirror')
        repository3 = self.create_repository(
            name='repo3',
            path='/path3',
            mirror_path='mirror')

        # This should fail, since all are visible and they conflict.
        with self.assertRaises(Repository.MultipleObjectsReturned):
            Repository.objects.get_best_match('mirror')

        # It should then work if only one is visible.
        repository2.visible = False
        repository2.save(update_fields=('visible',))

        repository3.visible = False
        repository3.save(update_fields=('visible',))

        self.assertEqual(
            Repository.objects.get_best_match('mirror'),
            repository1)

    def test_with_multiple_prefer_name(self):
        """Testing Repository.objects.get_best_match with multiple results
        prefers name over path/mirror_path
        """
        repository1 = self.create_repository(
            name='repo1',
            path='/path1',
            mirror_path='mirror')
        self.create_repository(
            name='repo2',
            path='/path2',
            mirror_path='mirror')
        self.create_repository(
            name='repo3',
            path='/path3',
            mirror_path='mirror')

        # This should fail, since all are visible and they conflict.
        with self.assertRaises(Repository.MultipleObjectsReturned):
            Repository.objects.get_best_match('mirror')

        # It should then work if only one is visible.
        repository1.name = 'mirror'
        repository1.save(update_fields=('name',))

        self.assertEqual(
            Repository.objects.get_best_match('mirror'),
            repository1)

    def test_with_multiple_prefer_path(self):
        """Testing Repository.objects.get_best_match with multiple results
        prefers path over mirror_path
        """
        repository1 = self.create_repository(
            name='repo1',
            path='/path1',
            mirror_path='mirror')
        self.create_repository(
            name='repo2',
            path='/path2',
            mirror_path='mirror')
        self.create_repository(
            name='repo3',
            path='/path3',
            mirror_path='mirror')

        # This should fail, since all are visible and they conflict.
        with self.assertRaises(Repository.MultipleObjectsReturned):
            Repository.objects.get_best_match('mirror')

        # It should then work if only one is visible.
        repository1.path = 'mirror'
        repository1.save(update_fields=('path',))

        self.assertEqual(
            Repository.objects.get_best_match('mirror'),
            repository1)



"""Unit tests for reviewboard.certs.errors.CertificateVerificationFailureCode.

Version Added:
    8.0
"""

from __future__ import annotations

from reviewboard.certs.errors import CertificateVerificationFailureCode
from reviewboard.certs.tests.testcases import CertificateTestCase


class CertificateVerificationFailureCodeTests(CertificateTestCase):
    """Unit tests for CertificateVerificationFailureCode.

    Version Added:
        8.0
    """

    def test_for_ssl_verify_code(self) -> None:
        """Testing CertificateVerificationError.for_ssl_verify_code"""
        # There's a bit of a tautology here with the test, but we're just
        # making sure there are no surprises in the future if behavior
        # changes.
        checks = [
            (9, CertificateVerificationFailureCode.NOT_YET_VALID),
            (10, CertificateVerificationFailureCode.EXPIRED),
            (18, CertificateVerificationFailureCode.NOT_TRUSTED),
            (19, CertificateVerificationFailureCode.NOT_TRUSTED),
            (20, CertificateVerificationFailureCode.NOT_TRUSTED),
            (21, CertificateVerificationFailureCode.NOT_TRUSTED),
            (27, CertificateVerificationFailureCode.NOT_TRUSTED),
            (62, CertificateVerificationFailureCode.HOSTNAME_MISMATCH),
            (100, CertificateVerificationFailureCode.OTHER),
            (0, CertificateVerificationFailureCode.OTHER),
            (-1, CertificateVerificationFailureCode.OTHER),
            (12345, CertificateVerificationFailureCode.OTHER),
        ]

        for ssl_code, expected_failure_code in checks:
            failure_code = (
                CertificateVerificationFailureCode
                .for_ssl_verify_code(ssl_code)
            )

            self.assertEqual(failure_code,
                             expected_failure_code,
                             f'SSL verification code {ssl_code}')
